import logging
import socket

from zeroconf import (
    EventLoopBlocked,
    ServiceInfo,
    Zeroconf,
)

logger = logging.getLogger("nanolib.discovery.mdns.broadcaster")


class PeerBroadcaster:
    """
    Broadcasts this peer's presence on the local network using mDNS/zeroconf.
    Registers a service with the peer's ID in the TXT record as per libp2p spec.
    """

    def __init__(
        self,
        zeroconf: Zeroconf,
        service_type: str,
        service_name: str,
        peer_id: str,
        port: int,
    ):
        self.zeroconf = zeroconf
        self.service_type = service_type
        self.peer_id = peer_id
        self.port = port
        self.service_name = service_name

        # Get the local IP address
        local_ip = self._get_local_ip()
        hostname = socket.gethostname()

        self.service_info = ServiceInfo(
            type_=self.service_type,
            name=self.service_name,
            port=self.port,
            properties={b"id": self.peer_id.encode()},
            server=f"{hostname}.local.",
            addresses=[socket.inet_aton(local_ip)],
        )

    def _get_local_ip(self) -> str:
        """Get the local IP address of this machine"""
        try:
            # Connect to a remote address to determine the local IP
            # This doesn't actually send data
            with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
                s.connect(("8.8.8.8", 80))
                local_ip = s.getsockname()[0]
            return local_ip
        except Exception:
            # Fallback to localhost if we can't determine the IP
            return "127.0.0.1"

    def register(self) -> None:
        """Register the peer's mDNS service on the network."""
        try:
            self.zeroconf.register_service(self.service_info)
            logger.debug(f"mDNS service registered: {self.service_name}")
        except EventLoopBlocked as e:
            logger.warning(
                "EventLoopBlocked while registering mDNS '%s': %s", self.service_name, e
            )
        except Exception as e:
            logger.error(
                "Unexpected error during mDNS registration for '%s': %r",
                self.service_name,
                e,
            )

    def unregister(self) -> None:
        """Unregister the peer's mDNS service from the network."""
        try:
            self.zeroconf.unregister_service(self.service_info)
            logger.debug(f"mDNS service unregistered: {self.service_name}")
        except EventLoopBlocked as e:
            logger.warning(
                "EventLoopBlocked while unregistering mDNS '%s': %s",
                self.service_name,
                e,
            )
        except Exception as e:
            logger.error(
                "Unexpected error during mDNS unregistration for '%s': %r",
                self.service_name,
                e,
            )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: libp2p/peer/persistent/pb/persistent_peerstore.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'libp2p/peer/persistent/pb/persistent_peerstore.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nanolib.peer.pb import crypto_pb2 as libp2p_dot_peer_dot_pb_dot_crypto__pb2
from nanolib.peer.pb import envelope_pb2 as libp2p_dot_peer_dot_pb_dot_envelope__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n4libp2p/peer/persistent/pb/persistent_peerstore.proto\x12\x19libp2p.peer.persistent.pb\x1a\x1blibp2p/peer/pb/crypto.proto\x1a\x1dlibp2p/peer/pb/envelope.proto\"\"\n\rPeerAddresses\x12\x11\n\taddresses\x18\x01 \x03(\x0c\"\"\n\rPeerProtocols\x12\x11\n\tprotocols\x18\x01 \x03(\t\"\x88\x01\n\x0cPeerMetadata\x12G\n\x08metadata\x18\x01 \x03(\x0b\x32\x35.libp2p.peer.persistent.pb.PeerMetadata.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\"x\n\x08PeerKeys\x12\x34\n\npublic_key\x18\x01 \x01(\x0b\x32 .libp2p.peer.pb.crypto.PublicKey\x12\x36\n\x0bprivate_key\x18\x02 \x01(\x0b\x32!.libp2p.peer.pb.crypto.PrivateKey\"!\n\x0bPeerLatency\x12\x12\n\nlatency_ns\x18\x01 \x01(\x03\"A\n\x0cPeerEnvelope\x12\x31\n\x08\x65nvelope\x18\x01 \x01(\x0b\x32\x1f.libp2p.peer.pb.record.Envelope\"\x8d\x01\n\x0fPeerRecordState\x12?\n\x05state\x18\x01 \x01(\x0e\x32\x30.libp2p.peer.persistent.pb.PeerRecordState.State\"9\n\x05State\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05VALID\x10\x01\x12\x0b\n\x07INVALID\x10\x02\x12\x0b\n\x07\x45XPIRED\x10\x03\x42\x37Z5github.com/libp2p/py-libp2p/libp2p/peer/persistent/pbb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nanolib.peer.persistent.pb.persistent_peerstore_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z5github.com/libp2p/py-libp2p/libp2p/peer/persistent/pb'
  _globals['_PEERMETADATA_METADATAENTRY']._loaded_options = None
  _globals['_PEERMETADATA_METADATAENTRY']._serialized_options = b'8\001'
  _globals['_PEERADDRESSES']._serialized_start=143
  _globals['_PEERADDRESSES']._serialized_end=177
  _globals['_PEERPROTOCOLS']._serialized_start=179
  _globals['_PEERPROTOCOLS']._serialized_end=213
  _globals['_PEERMETADATA']._serialized_start=216
  _globals['_PEERMETADATA']._serialized_end=352
  _globals['_PEERMETADATA_METADATAENTRY']._serialized_start=305
  _globals['_PEERMETADATA_METADATAENTRY']._serialized_end=352
  _globals['_PEERKEYS']._serialized_start=354
  _globals['_PEERKEYS']._serialized_end=474
  _globals['_PEERLATENCY']._serialized_start=476
  _globals['_PEERLATENCY']._serialized_end=509
  _globals['_PEERENVELOPE']._serialized_start=511
  _globals['_PEERENVELOPE']._serialized_end=576
  _globals['_PEERRECORDSTATE']._serialized_start=579
  _globals['_PEERRECORDSTATE']._serialized_end=720
  _globals['_PEERRECORDSTATE_STATE']._serialized_start=663
  _globals['_PEERRECORDSTATE_STATE']._serialized_end=720
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: libp2p/bitswap/pb/unixfs.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1elibp2p/bitswap/pb/unixfs.proto\x12\tunixfs.pb\"\x93\x02\n\x04\x44\x61ta\x12&\n\x04Type\x18\x01 \x02(\x0e\x32\x18.unixfs.pb.Data.DataType\x12\x0c\n\x04\x44\x61ta\x18\x02 \x01(\x0c\x12\x10\n\x08\x66ilesize\x18\x03 \x01(\x04\x12\x12\n\nblocksizes\x18\x04 \x03(\x04\x12\x10\n\x08hashType\x18\x05 \x01(\x04\x12\x0e\n\x06\x66\x61nout\x18\x06 \x01(\x04\x12\x0c\n\x04mode\x18\x07 \x01(\r\x12\'\n\x05mtime\x18\x08 \x01(\x0b\x32\x18.unixfs.pb.IPFSTimestamp\"V\n\x08\x44\x61taType\x12\x07\n\x03Raw\x10\x00\x12\r\n\tDirectory\x10\x01\x12\x08\n\x04\x46ile\x10\x02\x12\x0c\n\x08Metadata\x10\x03\x12\x0b\n\x07Symlink\x10\x04\x12\r\n\tHAMTShard\x10\x05\"\x1c\n\x08Metadata\x12\x10\n\x08MimeType\x18\x01 \x01(\t\"/\n\rIPFSTimestamp\x12\x0f\n\x07seconds\x18\x01 \x02(\x03\x12\r\n\x05nanos\x18\x02 \x01(\x07')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nanolib.bitswap.pb.unixfs_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_DATA']._serialized_start=46
  _globals['_DATA']._serialized_end=321
  _globals['_DATA_DATATYPE']._serialized_start=235
  _globals['_DATA_DATATYPE']._serialized_end=321
  _globals['_METADATA']._serialized_start=323
  _globals['_METADATA']._serialized_end=351
  _globals['_IPFSTIMESTAMP']._serialized_start=353
  _globals['_IPFSTIMESTAMP']._serialized_end=400
# @@protoc_insertion_point(module_scope)



"""
Kademlia DHT implementation for py-libp2p.

This module provides a Distributed Hash Table (DHT) implementation
based on the Kademlia protocol.
"""

from .kad_dht import (
    KadDHT,
)
from .peer_routing import (
    PeerRouting,
)
from .routing_table import (
    RoutingTable,
)
from .utils import (
    create_key_from_binary,
)
from .value_store import (
    ValueStore,
)

__all__ = [
    "KadDHT",
    "RoutingTable",
    "PeerRouting",
    "ValueStore",
    "create_key_from_binary",
]



"""
Ed25519 cryptographic key implementation.

This module provides Ed25519 public and private key implementations using
the PyNaCl library. Ed25519 is used for libp2p peer identity signatures.
"""

from nacl.exceptions import (
    BadSignatureError,
)
from nacl.signing import (
    SigningKey as PrivateKeyImpl,
    VerifyKey as PublicKeyImpl,
)

from nanolib.crypto.keys import (
    KeyPair,
    KeyType,
    PrivateKey,
    PublicKey,
)


class Ed25519PublicKey(PublicKey):
    """
    Ed25519 public key implementation.

    Provides cryptographic operations for Ed25519 public keys including
    signature verification. Used for libp2p peer identity verification.
    """

    def __init__(self, impl: PublicKeyImpl) -> None:
        self.impl = impl

    def to_bytes(self) -> bytes:
        return bytes(self.impl)

    @classmethod
    def from_bytes(cls, key_bytes: bytes) -> "Ed25519PublicKey":
        return cls(PublicKeyImpl(key_bytes))

    def get_type(self) -> KeyType:
        return KeyType.Ed25519

    def verify(self, data: bytes, signature: bytes) -> bool:
        try:
            self.impl.verify(data, signature)
        except BadSignatureError:
            return False
        return True


class Ed25519PrivateKey(PrivateKey):
    """
    Ed25519 private key implementation.

    Provides cryptographic operations for Ed25519 private keys including
    signature generation and public key derivation. Used for libp2p peer
    identity signing.
    """

    def __init__(self, impl: PrivateKeyImpl) -> None:
        self.impl = impl

    @classmethod
    def new(cls, seed: bytes | None = None) -> "Ed25519PrivateKey":
        if seed:
            private_key_impl = PrivateKeyImpl(seed)
        else:
            private_key_impl = PrivateKeyImpl.generate()
        return cls(private_key_impl)

    def to_bytes(self) -> bytes:
        return bytes(self.impl)

    @classmethod
    def from_bytes(cls, data: bytes) -> "Ed25519PrivateKey":
        impl = PrivateKeyImpl(data)
        return cls(impl)

    def get_type(self) -> KeyType:
        return KeyType.Ed25519

    def sign(self, data: bytes) -> bytes:
        signed = self.impl.sign(data)
        return signed.signature

    def get_public_key(self) -> PublicKey:
        return Ed25519PublicKey(self.impl.verify_key)


def create_new_key_pair(seed: bytes | None = None) -> KeyPair:
    private_key = Ed25519PrivateKey.new(seed)
    public_key = private_key.get_public_key()
    return KeyPair(private_key, public_key)



import time

from .base_timed_cache import (
    BaseTimedCache,
)


class FirstSeenCache(BaseTimedCache):
    """Cache where expiry is set only when first added."""

    def add(self, key: bytes) -> bool:
        now = int(time.time())
        with self.lock:
            if key in self.cache:
                # Check if the key is expired
                if self.cache[key] <= now:
                    # Key is expired, update the expiry and treat as a new entry
                    self.cache[key] = now + self.ttl
                    return True
                return False
            self.cache[key] = now + self.ttl
            return True

    def has(self, key: bytes) -> bool:
        now = int(time.time())
        with self.lock:
            if key in self.cache:
                # Check if key is expired
                if self.cache[key] <= now:
                    return False
                return True
            return False



"""NanoLib - P2P networking library for NanoPy blockchain (libp2p fork)."""

from __future__ import annotations

import logging
import ssl

from nanolib.transport.quic.utils import is_quic_multiaddr
from typing import Any
from nanolib.transport.quic.transport import QUICTransport
from nanolib.transport.quic.config import QUICTransportConfig
from collections.abc import (
    Mapping,
    Sequence,
)
from typing import (
    Literal,
    Optional,
)

import multiaddr

from nanolib.abc import (
    IHost,
    INetworkService,
    IPeerRouting,
    IPeerStore,
    ISecureTransport,
    ITransport,
)
from nanolib.rcmgr import ResourceManager
from nanolib.crypto.keys import (
    KeyPair,
)
from nanolib.crypto.ed25519 import (
    create_new_key_pair as create_new_ed25519_key_pair,
)
from nanolib.crypto.rsa import (
    create_new_key_pair as create_new_rsa_key_pair,
)
from nanolib.crypto.x25519 import create_new_key_pair as create_new_x25519_key_pair
from nanolib.custom_types import (
    TMuxerOptions,
    TProtocol,
    TSecurityOptions,
)
from nanolib.host.basic_host import (
    BasicHost,
)
from nanolib.host.routed_host import (
    RoutedHost,
)
from nanolib.network.swarm import (
    Swarm,
)
from nanolib.network.config import (
    ConnectionConfig,
    RetryConfig
)
from nanolib.peer.id import (
    ID,
)
from nanolib.peer.peerstore import (
    PeerStore,
    create_signed_peer_record,
)
from nanolib.peer.persistent import (
    create_sync_peerstore,
    create_async_peerstore,
    create_sync_sqlite_peerstore,
    create_async_sqlite_peerstore,
    create_sync_memory_peerstore,
    create_async_memory_peerstore,
    create_sync_leveldb_peerstore,
    create_async_leveldb_peerstore,
    create_sync_rocksdb_peerstore,
    create_async_rocksdb_peerstore,
)
from nanolib.security.insecure.transport import (
    PLAINTEXT_PROTOCOL_ID,
    InsecureTransport,
)
from nanolib.security.noise.transport import (
    PROTOCOL_ID as NOISE_PROTOCOL_ID,
    Transport as NoiseTransport,
)
import nanolib.security.secio.transport as secio
from nanolib.stream_muxer.mplex.mplex import (
    MPLEX_PROTOCOL_ID,
    Mplex,
)
from nanolib.stream_muxer.yamux.yamux import (
    PROTOCOL_ID as YAMUX_PROTOCOL_ID,
    Yamux,
)
from nanolib.transport.tcp.tcp import (
    TCP,
)
from nanolib.transport.upgrader import (
    TransportUpgrader,
)
from nanolib.transport.transport_registry import (
    create_transport_for_multiaddr,
    get_supported_transport_protocols,
)
from nanolib.utils.logging import (
    setup_logging,
)

# Initialize logging configuration
setup_logging()

# Default multiplexer choice
DEFAULT_MUXER = "YAMUX"

# Multiplexer options
MUXER_YAMUX = "YAMUX"
MUXER_MPLEX = "MPLEX"
DEFAULT_NEGOTIATE_TIMEOUT = 30  # seconds - increased for high-concurrency scenarios

logger = logging.getLogger(__name__)

def set_default_muxer(muxer_name: Literal["YAMUX", "MPLEX"]) -> None:
    """
    Set the default multiplexer protocol to use.

    :param muxer_name: Either "YAMUX" or "MPLEX"
    :raise ValueError: If an unsupported muxer name is provided
    """
    global DEFAULT_MUXER
    muxer_upper = muxer_name.upper()
    if muxer_upper not in [MUXER_YAMUX, MUXER_MPLEX]:
        raise ValueError(f"Unknown muxer: {muxer_name}. Use 'YAMUX' or 'MPLEX'.")
    DEFAULT_MUXER = muxer_upper


def get_default_muxer() -> str:
    """
    Returns the currently selected default muxer.

    :return: Either "YAMUX" or "MPLEX"
    """
    return DEFAULT_MUXER


def create_yamux_muxer_option() -> TMuxerOptions:
    """
    Returns muxer options with Yamux as the primary choice.

    :return: Muxer options with Yamux first
    """
    return {
        TProtocol(YAMUX_PROTOCOL_ID): Yamux,  # Primary choice
        TProtocol(MPLEX_PROTOCOL_ID): Mplex,  # Fallback for compatibility
    }


def create_mplex_muxer_option() -> TMuxerOptions:
    """
    Returns muxer options with Mplex as the primary choice.

    :return: Muxer options with Mplex first
    """
    return {
        TProtocol(MPLEX_PROTOCOL_ID): Mplex,  # Primary choice
        TProtocol(YAMUX_PROTOCOL_ID): Yamux,  # Fallback
    }


def generate_new_rsa_identity() -> KeyPair:
    return create_new_rsa_key_pair()


def generate_new_ed25519_identity() -> KeyPair:
    """
    Generate a new Ed25519 identity key pair.

    Ed25519 is preferred for better interoperability with other libp2p implementations
    (e.g., Rust, Go) which often disable RSA support.
    """
    return create_new_ed25519_key_pair()


def generate_peer_id_from(key_pair: KeyPair) -> ID:
    public_key = key_pair.public_key
    return ID.from_pubkey(public_key)


def get_default_muxer_options() -> TMuxerOptions:
    """
    Returns the default muxer options based on the current default muxer setting.

    :return: Muxer options with the preferred muxer first
    """
    if DEFAULT_MUXER == "MPLEX":
        return create_mplex_muxer_option()
    else:  # YAMUX is default
        return create_yamux_muxer_option()

def new_swarm(
    key_pair: KeyPair | None = None,
    muxer_opt: TMuxerOptions | None = None,
    sec_opt: TSecurityOptions | None = None,
    peerstore_opt: IPeerStore | None = None,
    muxer_preference: Literal["YAMUX", "MPLEX"] | None = None,
    listen_addrs: Sequence[multiaddr.Multiaddr] | None = None,
    enable_quic: bool = False,
    retry_config: RetryConfig | None = None,
    connection_config: ConnectionConfig | QUICTransportConfig | None = None,
    tls_client_config: ssl.SSLContext | None = None,
    tls_server_config: ssl.SSLContext | None = None,
    resource_manager: ResourceManager | None = None,
    psk: str | None = None
) -> INetworkService:
    logger.debug(f"new_swarm: enable_quic={enable_quic}, listen_addrs={listen_addrs}")
    """
    Create a swarm instance based on the parameters.

    :param key_pair: optional choice of the ``KeyPair``
    :param muxer_opt: optional choice of stream muxer
    :param sec_opt: optional choice of security upgrade
    :param peerstore_opt: optional peerstore
    :param muxer_preference: optional explicit muxer preference
    :param listen_addrs: optional list of multiaddrs to listen on
    :param enable_quic: enable quic for transport
    :param quic_transport_opt: options for transport
    :param resource_manager: optional resource manager for connection/stream limits
    :type resource_manager: :class:`libp2p.rcmgr.ResourceManager` or None
    :param psk: optional pre-shared key for PSK encryption in transport
    :return: return a default swarm instance

    Note: Yamux (/yamux/1.0.0) is the preferred stream multiplexer
          due to its improved performance and features.
          Mplex (/mplex/6.7.0) is retained for backward compatibility
          but may be deprecated in the future.

    Note: Ed25519 keys are used by default for better interoperability with
          other libp2p implementations (Rust, Go) which often disable RSA support.
    """
    if key_pair is None:
        # Use Ed25519 by default for better interoperability with Rust/Go libp2p
        # which often compile without RSA support
        key_pair = generate_new_ed25519_identity()

    id_opt = generate_peer_id_from(key_pair)

    transport: TCP | QUICTransport | ITransport
    quic_transport_opt = connection_config if isinstance(connection_config, QUICTransportConfig) else None

    if listen_addrs is None:
        if enable_quic:
            transport = QUICTransport(key_pair.private_key, config=quic_transport_opt)
        else:
            transport = TCP()
    else:
        # Use transport registry to select the appropriate transport
        from nanolib.transport.transport_registry import create_transport_for_multiaddr

        # Create a temporary upgrader for transport selection
        # We'll create the real upgrader later with the proper configuration
        temp_upgrader = TransportUpgrader(
            secure_transports_by_protocol={},
            muxer_transports_by_protocol={}
        )

        addr = listen_addrs[0]
        logger.debug(f"new_swarm: Creating transport for address: {addr}")
        transport_maybe = create_transport_for_multiaddr(
            addr,
            temp_upgrader,
            private_key=key_pair.private_key,
            config=quic_transport_opt,
            tls_client_config=tls_client_config,
            tls_server_config=tls_server_config
        )

        if transport_maybe is None:
            raise ValueError(f"Unsupported transport for listen_addrs: {listen_addrs}")

        transport = transport_maybe
        logger.debug(f"new_swarm: Created transport: {type(transport)}")

    # If enable_quic is True but we didn't get a QUIC transport, force QUIC
    if enable_quic and not isinstance(transport, QUICTransport):
        logger.debug(f"new_swarm: Forcing QUIC transport (enable_quic=True but got {type(transport)})")
        transport = QUICTransport(key_pair.private_key, config=quic_transport_opt)

    logger.debug(f"new_swarm: Final transport type: {type(transport)}")

    # Generate X25519 keypair for Noise
    noise_key_pair = create_new_x25519_key_pair()

    # Default security transports (using Noise as primary)
    secure_transports_by_protocol: Mapping[TProtocol, ISecureTransport] = sec_opt or {
        NOISE_PROTOCOL_ID: NoiseTransport(
            key_pair, noise_privkey=noise_key_pair.private_key
        ),
        TProtocol(secio.ID): secio.Transport(key_pair),
        TProtocol(PLAINTEXT_PROTOCOL_ID): InsecureTransport(
            key_pair, peerstore=peerstore_opt
        ),
    }

    # Use given muxer preference if provided, otherwise use global default
    if muxer_preference is not None:
        temp_pref = muxer_preference.upper()
        if temp_pref not in [MUXER_YAMUX, MUXER_MPLEX]:
            raise ValueError(
                f"Unknown muxer: {muxer_preference}. Use 'YAMUX' or 'MPLEX'."
            )
        active_preference = temp_pref
    else:
        active_preference = DEFAULT_MUXER

    # Use provided muxer options if given, otherwise create based on preference
    if muxer_opt is not None:
        muxer_transports_by_protocol = muxer_opt
    else:
        if active_preference == MUXER_MPLEX:
            muxer_transports_by_protocol = create_mplex_muxer_option()
        else:  # YAMUX is default
            muxer_transports_by_protocol = create_yamux_muxer_option()

    upgrader = TransportUpgrader(
        secure_transports_by_protocol=secure_transports_by_protocol,
        muxer_transports_by_protocol=muxer_transports_by_protocol,
    )


    peerstore = peerstore_opt or PeerStore()
    # Store our key pair in peerstore
    peerstore.add_key_pair(id_opt, key_pair)

    swarm = Swarm(
        id_opt,
        peerstore,
        upgrader,
        transport,
        retry_config=retry_config,
        connection_config=connection_config,
        psk=psk
    )

    # Set resource manager if provided
    # Auto-create a default ResourceManager if one was not provided
    if resource_manager is None:
        try:
            from nanolib.rcmgr import new_resource_manager as _new_rm

            resource_manager = _new_rm()
        except Exception:
            resource_manager = None

    if resource_manager is not None:
        swarm.set_resource_manager(resource_manager)

    return swarm

    # Set resource manager if provided
    # Auto-create a default ResourceManager if one was not provided
    if resource_manager is None:
        try:
            from nanolib.rcmgr import new_resource_manager as _new_rm

            resource_manager = _new_rm()
        except Exception:
            resource_manager = None

    if resource_manager is not None:
        swarm.set_resource_manager(resource_manager)

    return swarm


def new_host(
    key_pair: KeyPair | None = None,
    muxer_opt: TMuxerOptions | None = None,
    sec_opt: TSecurityOptions | None = None,
    peerstore_opt: IPeerStore | None = None,
    disc_opt: IPeerRouting | None = None,
    muxer_preference: Literal["YAMUX", "MPLEX"] | None = None,
    listen_addrs: Sequence[multiaddr.Multiaddr] | None = None,
    enable_mDNS: bool = False,
    enable_upnp: bool = False,
    bootstrap: list[str] | None = None,
    negotiate_timeout: int = DEFAULT_NEGOTIATE_TIMEOUT,
    enable_quic: bool = False,
    quic_transport_opt: QUICTransportConfig | None = None,
    tls_client_config: ssl.SSLContext | None = None,
    tls_server_config: ssl.SSLContext | None = None,
    resource_manager: ResourceManager | None = None,
    psk: str | None = None
) -> IHost:
    """
    Create a new libp2p host based on the given parameters.

    :param key_pair: optional choice of the ``KeyPair``
    :param muxer_opt: optional choice of stream muxer
    :param sec_opt: optional choice of security upgrade
    :param peerstore_opt: optional peerstore
    :param disc_opt: optional discovery
    :param muxer_preference: optional explicit muxer preference
    :param listen_addrs: optional list of multiaddrs to listen on
    :param enable_mDNS: whether to enable mDNS discovery
    :param bootstrap: optional list of bootstrap peer addresses as strings
    :param enable_quic: optinal choice to use QUIC for transport
    :param quic_transport_opt: optional configuration for quic transport
    :param tls_client_config: optional TLS client configuration for WebSocket transport
    :param tls_server_config: optional TLS server configuration for WebSocket transport
    :param resource_manager: optional resource manager for connection/stream limits
    :type resource_manager: :class:`libp2p.rcmgr.ResourceManager` or None
    :param psk: optional pre-shared key (PSK)
    :return: return a host instance
    """

    if not enable_quic and quic_transport_opt is not None:
        logger.warning(f"QUIC config provided but QUIC not enabled, ignoring QUIC config")

    # Enable automatic protection by default: if no resource manager is supplied,
    # create a default instance so connections/streams are guarded out of the box.
    if resource_manager is None:
        try:
            from nanolib.rcmgr import new_resource_manager as _new_rm

            resource_manager = _new_rm()
        except Exception:
            # Fallback to leaving it None if creation fails for any reason.
            resource_manager = None

    swarm = new_swarm(
        enable_quic=enable_quic,
        key_pair=key_pair,
        muxer_opt=muxer_opt,
        sec_opt=sec_opt,
        peerstore_opt=peerstore_opt,
        muxer_preference=muxer_preference,
        listen_addrs=listen_addrs,
        connection_config=quic_transport_opt if enable_quic else None,
        tls_client_config=tls_client_config,
        tls_server_config=tls_server_config,
        resource_manager=resource_manager,
        psk=psk
    )

    if disc_opt is not None:
        return RoutedHost(
            network=swarm,
            router=disc_opt,
            enable_mDNS=enable_mDNS,
            enable_upnp=enable_upnp,
            bootstrap=bootstrap,
            resource_manager=resource_manager,
        )
    return BasicHost(
        network=swarm,
        enable_mDNS=enable_mDNS,
        bootstrap=bootstrap,
        enable_upnp=enable_upnp,
        negotiate_timeout=negotiate_timeout,
        resource_manager=resource_manager,
    )

__version__ = "0.1.0"  # NanoLib version



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class Propose(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RAND_FIELD_NUMBER: builtins.int
    PUBLIC_KEY_FIELD_NUMBER: builtins.int
    EXCHANGES_FIELD_NUMBER: builtins.int
    CIPHERS_FIELD_NUMBER: builtins.int
    HASHES_FIELD_NUMBER: builtins.int
    rand: builtins.bytes
    public_key: builtins.bytes
    exchanges: builtins.str
    ciphers: builtins.str
    hashes: builtins.str
    def __init__(
        self,
        *,
        rand: builtins.bytes | None = ...,
        public_key: builtins.bytes | None = ...,
        exchanges: builtins.str | None = ...,
        ciphers: builtins.str | None = ...,
        hashes: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["ciphers", b"ciphers", "exchanges", b"exchanges", "hashes", b"hashes", "public_key", b"public_key", "rand", b"rand"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["ciphers", b"ciphers", "exchanges", b"exchanges", "hashes", b"hashes", "public_key", b"public_key", "rand", b"rand"]) -> None: ...

global___Propose = Propose

@typing.final
class Exchange(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    EPHEMERAL_PUBLIC_KEY_FIELD_NUMBER: builtins.int
    SIGNATURE_FIELD_NUMBER: builtins.int
    ephemeral_public_key: builtins.bytes
    signature: builtins.bytes
    def __init__(
        self,
        *,
        ephemeral_public_key: builtins.bytes | None = ...,
        signature: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["ephemeral_public_key", b"ephemeral_public_key", "signature", b"signature"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["ephemeral_public_key", b"ephemeral_public_key", "signature", b"signature"]) -> None: ...

global___Exchange = Exchange



"""
NanoPy Proof of Stake Consensus

Supports both in-memory validators and on-chain staking contract.
"""

import time
import random
import hashlib
from typing import Dict, List, Optional, Tuple
from dataclasses import dataclass, field
from eth_utils import to_hex, keccak, to_bytes


@dataclass
class Validator:
    address: str
    stake: int
    registered_at: int
    last_block: int = 0
    blocks_validated: int = 0
    is_active: bool = True


@dataclass
class StakeInfo:
    amount: int
    locked_until: int
    validator: str


# Staking contract function selectors
STAKING_SELECTORS = {
    'getValidators': '0xb7ab4db5',      # getValidators() returns address[]
    'isValidator': '0xfacd743b',         # isValidator(address) returns bool
    'getStake': '0x7a766460',           # getStake(address) returns uint256
    'getValidatorCount': '0x7071688a',  # getValidatorCount() returns uint256
}


class ProofOfStake:
    """
    Proof of Stake consensus engine for NanoPy

    - Validators stake NPY to participate
    - Selection weighted by stake amount
    - Rewards distributed to validators
    - Supports on-chain staking contract
    - Bootstrap mode: Founder can validate before staking contract is deployed
    """

    MIN_STAKE = 10_000 * 10**18  # 10,000 NPY
    BLOCK_REWARD = 2 * 10**18    # 2 NPY per block
    LOCK_PERIOD = 100            # Blocks before unstake

    # Bootstrap validator (PoA) - can validate until staking contract is deployed
    BOOTSTRAP_VALIDATOR = "0x311384E7e4803AdBDB1D8ab46b973eeE0f9cF55B"

    def __init__(self, state=None, staking_contract: str = None, node=None):
        self.state = state
        self.staking_contract = staking_contract  # Contract address
        self.node = node  # Reference to node for contract calls
        self.validators: Dict[str, Validator] = {}
        self.stakes: Dict[str, StakeInfo] = {}
        self.current_validator: Optional[str] = None
        self.last_block_time = 0
        self.block_time = 12  # seconds
        self._contract_validators_cache: List[str] = []
        self._cache_block: int = -1

    def register_validator(self, address: str, stake: int, block_number: int) -> bool:
        """Register a new validator with stake"""
        if stake < self.MIN_STAKE:
            return False

        if address in self.validators:
            return False

        self.validators[address] = Validator(
            address=address,
            stake=stake,
            registered_at=block_number,
        )

        self.stakes[address] = StakeInfo(
            amount=stake,
            locked_until=block_number + self.LOCK_PERIOD,
            validator=address,
        )

        return True

    def add_stake(self, address: str, amount: int, block_number: int) -> bool:
        """Add more stake to existing validator"""
        if address not in self.validators:
            return False

        self.validators[address].stake += amount
        self.stakes[address].amount += amount
        self.stakes[address].locked_until = block_number + self.LOCK_PERIOD

        return True

    def withdraw_stake(self, address: str, block_number: int) -> Tuple[bool, int]:
        """Withdraw stake if lock period passed"""
        if address not in self.validators:
            return False, 0

        stake_info = self.stakes.get(address)
        if not stake_info:
            return False, 0

        if block_number < stake_info.locked_until:
            return False, 0

        amount = stake_info.amount

        del self.validators[address]
        del self.stakes[address]

        return True, amount

    def select_validator(self, block_number: int, block_hash: bytes) -> Optional[str]:
        """Select next validator weighted by stake"""
        active_validators = [v for v in self.validators.values() if v.is_active]

        # Bootstrap mode: if no validators registered, use bootstrap validator
        if not active_validators:
            if self._is_bootstrap_mode():
                return self.BOOTSTRAP_VALIDATOR
            return None

        total_stake = sum(v.stake for v in active_validators)
        if total_stake == 0:
            return None

        # Deterministic random based on block hash
        seed = int.from_bytes(
            keccak(block_hash + block_number.to_bytes(8, 'big'))[:8],
            'big'
        )
        random.seed(seed)

        # Weighted selection
        selection = random.randint(0, total_stake - 1)
        cumulative = 0

        for validator in active_validators:
            cumulative += validator.stake
            if selection < cumulative:
                self.current_validator = validator.address
                return validator.address

        return active_validators[-1].address

    def validate_block(self, block, validator_address: str) -> bool:
        """Verify block was created by valid validator"""
        if validator_address not in self.validators:
            return False

        validator = self.validators[validator_address]
        if not validator.is_active:
            return False

        # Check validator was selected for this block
        expected = self.select_validator(block.number - 1, block.parent_hash)
        if expected != validator_address:
            return False

        return True

    def apply_reward(self, validator_address: str, block_number: int):
        """Apply block reward to validator"""
        if validator_address not in self.validators:
            return

        validator = self.validators[validator_address]
        validator.last_block = block_number
        validator.blocks_validated += 1

        # Add reward to stake (compound)
        validator.stake += self.BLOCK_REWARD
        if validator_address in self.stakes:
            self.stakes[validator_address].amount += self.BLOCK_REWARD

    def slash_validator(self, address: str, reason: str = ""):
        """Slash validator for misbehavior"""
        if address not in self.validators:
            return

        validator = self.validators[address]

        # Slash 10% of stake
        slash_amount = validator.stake // 10
        validator.stake -= slash_amount

        if address in self.stakes:
            self.stakes[address].amount -= slash_amount

        # Deactivate if stake below minimum
        if validator.stake < self.MIN_STAKE:
            validator.is_active = False

    def get_validator_set(self) -> List[dict]:
        """Get list of all validators"""
        return [
            {
                "address": v.address,
                "stake": hex(v.stake),
                "blocks_validated": v.blocks_validated,
                "is_active": v.is_active,
            }
            for v in self.validators.values()
        ]

    def get_total_stake(self) -> int:
        """Get total staked amount"""
        return sum(v.stake for v in self.validators.values())

    def is_validator(self, address: str) -> bool:
        """Check if address is registered validator"""
        # Bootstrap mode: Founder can validate before staking contract exists
        if self._is_bootstrap_mode() and address.lower() == self.BOOTSTRAP_VALIDATOR.lower():
            return True
        # Check on-chain contract first
        if self.staking_contract and self.node:
            return self._is_validator_from_contract(address)
        # Fallback to in-memory
        return address in self.validators and self.validators[address].is_active

    def _is_bootstrap_mode(self) -> bool:
        """Check if we're in bootstrap mode (no staking contract deployed yet)"""
        if not self.staking_contract or not self.node:
            return True
        # Check if staking contract has code
        try:
            code = self.node.get_code(self.staking_contract)
            return code is None or code == b'' or code == '0x'
        except Exception:
            return True

    def get_stake(self, address: str) -> int:
        """Get stake amount for address"""
        # Check on-chain contract first
        if self.staking_contract and self.node:
            return self._get_stake_from_contract(address)
        # Fallback to in-memory
        if address in self.stakes:
            return self.stakes[address].amount
        return 0

    # Contract interaction methods
    def _call_contract(self, data: str) -> str:
        """Call staking contract"""
        if not self.node or not self.staking_contract:
            return "0x"
        try:
            return self.node.call({
                "to": self.staking_contract,
                "data": data,
            })
        except Exception:
            return "0x"

    def _is_validator_from_contract(self, address: str) -> bool:
        """Check isValidator(address) on contract"""
        # Encode: selector + address padded to 32 bytes
        addr = address.lower().replace("0x", "").zfill(64)
        data = STAKING_SELECTORS['isValidator'] + addr
        result = self._call_contract(data)
        if result and len(result) >= 66:
            return int(result[-1], 16) == 1
        return False

    def _get_stake_from_contract(self, address: str) -> int:
        """Get stake amount from contract"""
        addr = address.lower().replace("0x", "").zfill(64)
        data = STAKING_SELECTORS['getStake'] + addr
        result = self._call_contract(data)
        if result and len(result) > 2:
            return int(result, 16)
        return 0

    def _get_validators_from_contract(self) -> List[str]:
        """Get list of validators from contract"""
        data = STAKING_SELECTORS['getValidators']
        result = self._call_contract(data)
        if not result or result == "0x":
            return []
        # Decode dynamic array of addresses
        try:
            hex_data = result[2:]  # Remove 0x
            if len(hex_data) < 128:  # At least offset + length
                return []
            # First 32 bytes = offset to array data
            # Next 32 bytes = array length
            length = int(hex_data[64:128], 16)
            validators = []
            for i in range(length):
                start = 128 + i * 64
                addr_hex = hex_data[start:start+64]
                validators.append("0x" + addr_hex[-40:])
            return validators
        except Exception:
            return []

    def sync_from_contract(self):
        """Sync validator list from contract"""
        if not self.staking_contract or not self.node:
            return
        validators = self._get_validators_from_contract()
        for addr in validators:
            if addr.lower() not in [v.lower() for v in self.validators]:
                stake = self._get_stake_from_contract(addr)
                if stake >= self.MIN_STAKE:
                    self.validators[addr.lower()] = Validator(
                        address=addr,
                        stake=stake,
                        registered_at=0,
                        is_active=True
                    )



"""
Synchronous LevelDB datastore implementation for persistent peer storage.

This provides a synchronous LevelDB-based datastore for high-performance
persistent storage. LevelDB is a fast key-value storage library written at Google.
"""

from collections.abc import Iterator
import importlib
from pathlib import Path
import threading
from typing import Any

from .base_sync import IBatchingDatastoreSync, IBatchSync


class LevelDBBatchSync(IBatchSync):
    """Synchronous LevelDB batch implementation."""

    def __init__(self, db: "LevelDBDatastoreSync"):
        self.db = db
        self.operations: list[tuple[str, bytes, bytes | None]] = []

    def put(self, key: bytes, value: bytes) -> None:
        """Add a put operation to the batch."""
        self.operations.append(("put", key, value))

    def delete(self, key: bytes) -> None:
        """Add a delete operation to the batch."""
        self.operations.append(("delete", key, None))

    def commit(self) -> None:
        """Commit all operations in the batch."""
        with self.db._lock:
            try:
                # Create a write batch
                db = self.db.db
                if db is None:
                    raise ValueError("LevelDB database is not initialized")
                write_batch = db.WriteBatch()

                for operation, key, value in self.operations:
                    if operation == "put":
                        write_batch.Put(key, value)
                    elif operation == "delete":
                        write_batch.Delete(key)

                # Write the batch atomically
                db.Write(write_batch)
            except Exception as e:
                raise e
            finally:
                self.operations.clear()


class LevelDBDatastoreSync(IBatchingDatastoreSync):
    """
    Synchronous LevelDB-based datastore implementation.

    This provides high-performance persistent storage using LevelDB with
    synchronous operations.
    """

    def __init__(self, path: str | Path):
        """
        Initialize synchronous LevelDB datastore.

        Args:
            path: Path to the LevelDB database directory

        """
        self.path = Path(path)
        self.db: Any = None
        self._lock = threading.Lock()
        self._leveldb: Any = None
        self._ensure_connection()

    def _ensure_connection(self) -> None:
        """Ensure database connection is established."""
        if self.db is None:
            with self._lock:
                if self.db is None:
                    try:
                        # Import LevelDB (plyvel)
                        self._leveldb = importlib.import_module("plyvel")
                    except ImportError as e:
                        raise ImportError(
                            "LevelDB support requires 'plyvel' package. "
                            "Install it with: pip install plyvel"
                        ) from e

                    # Create directory if it doesn't exist
                    self.path.mkdir(parents=True, exist_ok=True)

                    # Open LevelDB database
                    self.db = self._leveldb.DB(str(self.path), create_if_missing=True)

    def get(self, key: bytes) -> bytes | None:
        """Retrieve a value by key."""
        self._ensure_connection()
        if self.db is None:
            raise ValueError("LevelDB database is not initialized")

        with self._lock:
            try:
                return self.db.Get(key)
            except KeyError:
                return None

    def put(self, key: bytes, value: bytes) -> None:
        """Store a key-value pair."""
        self._ensure_connection()
        if self.db is None:
            raise ValueError("LevelDB database is not initialized")

        with self._lock:
            self.db.Put(key, value)

    def delete(self, key: bytes) -> None:
        """Delete a key-value pair."""
        self._ensure_connection()
        if self.db is None:
            raise ValueError("LevelDB database is not initialized")

        with self._lock:
            try:
                self.db.Delete(key)
            except KeyError:
                pass  # Key doesn't exist, which is fine

    def has(self, key: bytes) -> bool:
        """Check if a key exists."""
        self._ensure_connection()
        if self.db is None:
            raise ValueError("LevelDB database is not initialized")

        with self._lock:
            try:
                self.db.Get(key)
                return True
            except KeyError:
                return False

    def query(self, prefix: bytes = b"") -> Iterator[tuple[bytes, bytes]]:
        """
        Query key-value pairs with optional prefix.
        """
        self._ensure_connection()
        if self.db is None:
            raise ValueError("LevelDB database is not initialized")

        with self._lock:
            # Create iterator with prefix
            if prefix:
                iterator = self.db.iterator(prefix=prefix)
            else:
                iterator = self.db.iterator()

            try:
                yield from iterator
            finally:
                iterator.close()

    def batch(self) -> IBatchSync:
        """Create a new batch for atomic operations."""
        self._ensure_connection()
        if self.db is None:
            raise ValueError("LevelDB database is not initialized")
        return LevelDBBatchSync(self)

    def sync(self, prefix: bytes) -> None:
        """Flush pending writes to disk."""
        # LevelDB automatically syncs writes, but we can force a sync
        if self.db is not None:
            with self._lock:
                # LevelDB doesn't have an explicit sync method in plyvel
                # Writes are automatically synced to disk
                pass

    def close(self) -> None:
        """Close the datastore connection."""
        with self._lock:
            if self.db:
                self.db.close()
                self.db = None

    def __enter__(self) -> "LevelDBDatastoreSync":
        """Context manager entry."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        """Context manager exit."""
        self.close()



"""
Cross-platform path utilities for py-libp2p.

This module provides standardized path operations to ensure consistent
behavior across Windows, macOS, and Linux platforms.
"""

import os
from pathlib import Path
import sys
import tempfile
from typing import Union

PathLike = Union[str, Path]


def get_temp_dir() -> Path:
    """
    Get cross-platform temporary directory.

    Returns:
        Path: Platform-specific temporary directory path

    """
    return Path(tempfile.gettempdir())


def get_project_root() -> Path:
    """
    Get the project root directory.

    Returns:
        Path: Path to the py-libp2p project root

    """
    # Navigate from libp2p/utils/paths.py to project root
    return Path(__file__).parent.parent.parent


def join_paths(*parts: PathLike) -> Path:
    """
    Cross-platform path joining.

    Args:
        *parts: Path components to join

    Returns:
        Path: Joined path using platform-appropriate separator

    """
    return Path(*parts)


def ensure_dir_exists(path: PathLike) -> Path:
    """
    Ensure directory exists, create if needed.

    Args:
        path: Directory path to ensure exists

    Returns:
        Path: Path object for the directory

    """
    path_obj = Path(path)
    path_obj.mkdir(parents=True, exist_ok=True)
    return path_obj


def get_config_dir() -> Path:
    """
    Get user config directory (cross-platform).

    Returns:
        Path: Platform-specific config directory

    """
    if os.name == "nt":  # Windows
        appdata = os.environ.get("APPDATA", "")
        if appdata:
            return Path(appdata) / "py-libp2p"
        else:
            # Fallback to user home directory
            return Path.home() / "AppData" / "Roaming" / "py-libp2p"
    else:  # Unix-like (Linux, macOS)
        return Path.home() / ".config" / "py-libp2p"


def get_script_dir(script_path: PathLike | None = None) -> Path:
    """
    Get the directory containing a script file.

    Args:
        script_path: Path to the script file. If None, uses __file__

    Returns:
        Path: Directory containing the script

    Raises:
        RuntimeError: If script path cannot be determined

    """
    if script_path is None:
        # This will be the directory of the calling script
        import inspect

        frame = inspect.currentframe()
        if frame and frame.f_back:
            script_path = frame.f_back.f_globals.get("__file__")
        else:
            raise RuntimeError("Could not determine script path")

    if script_path is None:
        raise RuntimeError("Script path is None")

    return Path(script_path).parent.absolute()


def create_temp_file(prefix: str = "py-libp2p_", suffix: str = ".log") -> Path:
    """
    Create a temporary file with a unique name.

    Args:
        prefix: File name prefix
        suffix: File name suffix

    Returns:
        Path: Path to the created temporary file

    """
    temp_dir = get_temp_dir()
    # Create a unique filename using timestamp and random bytes
    import secrets
    import time

    timestamp = time.strftime("%Y%m%d_%H%M%S")
    microseconds = f"{time.time() % 1:.6f}"[2:]  # Get microseconds as string
    unique_id = secrets.token_hex(4)
    filename = f"{prefix}{timestamp}_{microseconds}_{unique_id}{suffix}"

    temp_file = temp_dir / filename
    # Create the file by touching it
    temp_file.touch()
    return temp_file


def resolve_relative_path(base_path: PathLike, relative_path: PathLike) -> Path:
    """
    Resolve a relative path from a base path.

    Args:
        base_path: Base directory path
        relative_path: Relative path to resolve

    Returns:
        Path: Resolved absolute path

    """
    base = Path(base_path).resolve()
    relative = Path(relative_path)

    if relative.is_absolute():
        return relative
    else:
        return (base / relative).resolve()


def normalize_path(path: PathLike) -> Path:
    """
    Normalize a path, resolving any symbolic links and relative components.

    Args:
        path: Path to normalize

    Returns:
        Path: Normalized absolute path

    """
    return Path(path).resolve()


def get_venv_path() -> Path | None:
    """
    Get virtual environment path if active.

    Returns:
        Path: Virtual environment path if active, None otherwise

    """
    venv_path = os.environ.get("VIRTUAL_ENV")
    if venv_path:
        return Path(venv_path)
    return None


def get_python_executable() -> Path:
    """
    Get current Python executable path.

    Returns:
        Path: Path to the current Python executable

    """
    return Path(sys.executable)


def find_executable(name: str) -> Path | None:
    """
    Find executable in system PATH.

    Args:
        name: Name of the executable to find

    Returns:
        Path: Path to executable if found, None otherwise

    """
    # Check if name already contains path
    if os.path.dirname(name):
        path = Path(name)
        if path.exists() and os.access(path, os.X_OK):
            return path
        return None

    # Search in PATH
    for path_dir in os.environ.get("PATH", "").split(os.pathsep):
        if not path_dir:
            continue
        path = Path(path_dir) / name
        if path.exists() and os.access(path, os.X_OK):
            return path

    return None


def get_script_binary_path() -> Path:
    """
    Get path to script's binary directory.

    Returns:
        Path: Directory containing the script's binary

    """
    return get_python_executable().parent


def get_binary_path(binary_name: str) -> Path | None:
    """
    Find binary in PATH or virtual environment.

    Args:
        binary_name: Name of the binary to find

    Returns:
        Path: Path to binary if found, None otherwise

    """
    # First check in virtual environment if active
    venv_path = get_venv_path()
    if venv_path:
        venv_bin = venv_path / "bin" if os.name != "nt" else venv_path / "Scripts"
        binary_path = venv_bin / binary_name
        if binary_path.exists() and os.access(binary_path, os.X_OK):
            return binary_path

    # Fall back to system PATH
    return find_executable(binary_name)



"""
Rendezvous service implementation for hosting a rendezvous point.
"""

import logging
import time

import varint

from nanolib.abc import IHost, INetStream
from nanolib.peer.id import ID as PeerID

from .config import (
    MAX_DISCOVER_LIMIT,
    MAX_NAMESPACE_LENGTH,
    MAX_PEER_ADDRESS_LENGTH,
    MAX_REGISTRATIONS,
    MAX_TTL,
    RENDEZVOUS_PROTOCOL,
)
from .messages import (
    create_discover_response_message,
    create_register_response_message,
    parse_peer_info,
)
from .pb.rendezvous_pb2 import Message

logger = logging.getLogger(__name__)


class RegistrationRecord:
    """Represents a peer registration record."""

    def __init__(self, peer_id: PeerID, addrs: list[bytes], namespace: str, ttl: int):
        self.peer_id = peer_id
        self.addrs = addrs
        self.namespace = namespace
        self.ttl = ttl
        self.registered_at = time.time()
        self.expires_at = self.registered_at + ttl

    def is_expired(self) -> bool:
        """Check if this registration has expired."""
        return time.time() > self.expires_at

    def to_protobuf_register(self) -> Message.Register:
        """Convert to protobuf Register message."""
        register = Message.Register()
        register.ns = self.namespace
        register.peer.id = self.peer_id.to_bytes()
        register.peer.addrs.extend(self.addrs)
        register.ttl = max(0, int(self.expires_at - time.time()))
        return register


class RendezvousService:
    """
    Rendezvous service for hosting a rendezvous point.

    This service allows peers to register under namespaces and discover
    other peers that have registered under the same namespaces.
    """

    def __init__(self, host: IHost):
        """
        Initialize rendezvous service.

        Args:
            host: The libp2p host

        """
        self.host = host
        # Store registrations by namespace
        self.registrations: dict[str, dict[PeerID, RegistrationRecord]] = {}

        # Set up stream handler
        host.set_stream_handler(RENDEZVOUS_PROTOCOL, self._handle_stream)

        logger.info("Rendezvous service started")

    async def _handle_stream(self, stream: INetStream) -> None:
        """Handle incoming rendezvous protocol streams."""
        peer_id = stream.muxed_conn.peer_id
        logger.debug(f"New rendezvous stream from {peer_id}")

        try:
            while True:
                # Read message length
                length_bytes = b""
                while True:
                    b = await stream.read(1)
                    if not b:
                        return  # Stream closed
                    length_bytes += b
                    if b[0] & 0x80 == 0:
                        break

                message_length = varint.decode_bytes(length_bytes)

                # Read message data
                message_bytes = b""
                remaining = message_length
                while remaining > 0:
                    chunk = await stream.read(remaining)
                    if not chunk:
                        return  # Stream closed
                    message_bytes += chunk
                    remaining -= len(chunk)

                # Parse message
                request = Message()
                request.ParseFromString(message_bytes)

                # Handle message based on type
                response = None
                if request.type == Message.REGISTER:
                    response = self._handle_register(peer_id, request.register)
                elif request.type == Message.UNREGISTER:
                    self._handle_unregister(peer_id, request.unregister)
                    # No response for unregister
                elif request.type == Message.DISCOVER:
                    response = self._handle_discover(peer_id, request.discover)
                else:
                    logger.warning(f"Unknown message type: {request.type}")
                    return

                # Send response if needed
                if response:
                    response_bytes = response.SerializeToString()
                    await stream.write(varint.encode(len(response_bytes)))
                    await stream.write(response_bytes)

        except Exception as e:
            logger.error(f"Error handling stream from {peer_id}: {e}")
        finally:
            await stream.close()

    def _handle_register(
        self, peer_id: PeerID, register_msg: Message.Register
    ) -> Message:
        """Handle REGISTER message."""
        target_peer_id = PeerID(register_msg.peer.id)
        namespace = register_msg.ns
        ttl = register_msg.ttl

        # Only allow peers to register themselves
        if peer_id != target_peer_id:
            logger.warning(
                f"Peer {peer_id} tried to register {target_peer_id} "
                f"in namespace '{namespace}'"
            )
            return create_register_response_message(
                Message.ResponseStatus.E_NOT_AUTHORIZED, "Peer can only register itself"
            )

        # Validate namespace
        if not namespace or len(namespace) > MAX_NAMESPACE_LENGTH:
            return create_register_response_message(
                Message.ResponseStatus.E_INVALID_NAMESPACE, "Invalid namespace"
            )

        # Validate TTL
        if ttl <= 0 or ttl > MAX_TTL:
            return create_register_response_message(
                Message.ResponseStatus.E_INVALID_TTL,
                f"TTL must be between 1 and {MAX_TTL} seconds",
            )

        # Validate peer info
        if not register_msg.peer.id:
            return create_register_response_message(
                Message.ResponseStatus.E_INVALID_PEER_INFO, "Missing peer ID"
            )

        # Check address lengths
        for addr in register_msg.peer.addrs:
            if len(addr) > MAX_PEER_ADDRESS_LENGTH:
                return create_register_response_message(
                    Message.ResponseStatus.E_INVALID_PEER_INFO, "Address too long"
                )

        # Ensure namespace exists in registrations
        if namespace not in self.registrations:
            self.registrations[namespace] = {}

        # Check registration limit for namespace
        if len(self.registrations[namespace]) >= MAX_REGISTRATIONS:
            # Remove expired registrations first
            self._cleanup_expired_registrations(namespace)

            if len(self.registrations[namespace]) >= MAX_REGISTRATIONS:
                return create_register_response_message(
                    Message.ResponseStatus.E_UNAVAILABLE, "Registration limit reached"
                )

        # Create registration record
        try:
            reg_peer_id, _ = parse_peer_info(register_msg.peer)
        except Exception:
            return create_register_response_message(
                Message.ResponseStatus.E_INVALID_PEER_INFO, "Invalid peer info"
            )

        record = RegistrationRecord(
            reg_peer_id, list(register_msg.peer.addrs), namespace, ttl
        )

        # Store registration
        self.registrations[namespace][reg_peer_id] = record

        logger.info(
            f"Registered peer {reg_peer_id} in namespace '{namespace}' with TTL {ttl}s"
        )

        return create_register_response_message(Message.ResponseStatus.OK, "OK", ttl)

    def _handle_unregister(
        self, peer_id: PeerID, unregister_msg: Message.Unregister
    ) -> None:
        """Handle UNREGISTER message."""
        namespace = unregister_msg.ns
        target_peer_id = PeerID(unregister_msg.id)

        # Only allow peers to unregister themselves
        if peer_id != target_peer_id:
            logger.warning(
                f"Peer {peer_id} tried to unregister {target_peer_id} "
                f"from namespace '{namespace}'"
            )
            return

        # Remove registration
        if namespace in self.registrations:
            self.registrations[namespace].pop(target_peer_id, None)
            logger.info(
                f"Unregistered peer {target_peer_id} from namespace '{namespace}'"
            )

    def _handle_discover(
        self, peer_id: PeerID, discover_msg: Message.Discover
    ) -> Message:
        """Handle DISCOVER message."""
        namespace = discover_msg.ns
        limit = discover_msg.limit
        cookie = discover_msg.cookie

        # Validate namespace
        if not namespace or len(namespace) > MAX_NAMESPACE_LENGTH:
            return create_discover_response_message(
                [], b"", Message.ResponseStatus.E_INVALID_NAMESPACE, "Invalid namespace"
            )

        # Validate limit
        if limit <= 0 or limit > MAX_DISCOVER_LIMIT:
            limit = MAX_DISCOVER_LIMIT

        # Clean up expired registrations
        if namespace in self.registrations:
            self._cleanup_expired_registrations(namespace)
        else:
            self.registrations[namespace] = {}

        # Get registrations for namespace
        registrations = list(self.registrations[namespace].values())

        # Simple pagination using cookie as offset
        offset = 0
        if cookie:
            try:
                offset = int.from_bytes(cookie, "big")
            except (ValueError, OverflowError):
                return create_discover_response_message(
                    [], b"", Message.ResponseStatus.E_INVALID_COOKIE, "Invalid cookie"
                )

        # Get slice of registrations
        end_offset = min(offset + limit, len(registrations))
        slice_registrations = registrations[offset:end_offset]

        # Create new cookie for next page
        new_cookie = b""
        if end_offset < len(registrations):
            new_cookie = end_offset.to_bytes(4, "big")

        # Convert to protobuf Register messages
        pb_registrations = [reg.to_protobuf_register() for reg in slice_registrations]

        logger.debug(
            f"Discovered {len(pb_registrations)} peers in namespace '{namespace}' "
            f"for peer {peer_id}"
        )

        return create_discover_response_message(
            pb_registrations, new_cookie, Message.ResponseStatus.OK, "OK"
        )

    def _cleanup_expired_registrations(self, namespace: str) -> None:
        """Remove expired registrations from a namespace."""
        if namespace not in self.registrations:
            return

        expired_peers = [
            peer_id
            for peer_id, record in self.registrations[namespace].items()
            if record.is_expired()
        ]

        for peer_id in expired_peers:
            del self.registrations[namespace][peer_id]

        if expired_peers:
            logger.debug(
                f"Cleaned up {len(expired_peers)} expired registrations"
                f"from '{namespace}'"
            )

    def get_namespace_stats(self) -> dict[str, int]:
        """Get statistics about registrations per namespace."""
        stats = {}
        for namespace, registrations in self.registrations.items():
            # Clean up expired first
            self._cleanup_expired_registrations(namespace)
            stats[namespace] = len(registrations)
        return stats

    def cleanup_all_expired(self) -> None:
        """Clean up expired registrations from all namespaces."""
        for namespace in list(self.registrations.keys()):
            self._cleanup_expired_registrations(namespace)
            # Remove empty namespaces
            if not self.registrations[namespace]:
                del self.registrations[namespace]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Modified from https://github.com/libp2p/go-libp2p-pubsub/blob/master/pb/rpc.proto"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class RPC(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class SubOpts(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        SUBSCRIBE_FIELD_NUMBER: builtins.int
        TOPICID_FIELD_NUMBER: builtins.int
        subscribe: builtins.bool
        """subscribe or unsubscribe"""
        topicid: builtins.str
        def __init__(
            self,
            *,
            subscribe: builtins.bool | None = ...,
            topicid: builtins.str | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["subscribe", b"subscribe", "topicid", b"topicid"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["subscribe", b"subscribe", "topicid", b"topicid"]) -> None: ...

    SUBSCRIPTIONS_FIELD_NUMBER: builtins.int
    PUBLISH_FIELD_NUMBER: builtins.int
    CONTROL_FIELD_NUMBER: builtins.int
    SENDERRECORD_FIELD_NUMBER: builtins.int
    senderRecord: builtins.bytes
    @property
    def subscriptions(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___RPC.SubOpts]: ...
    @property
    def publish(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Message]: ...
    @property
    def control(self) -> global___ControlMessage: ...
    def __init__(
        self,
        *,
        subscriptions: collections.abc.Iterable[global___RPC.SubOpts] | None = ...,
        publish: collections.abc.Iterable[global___Message] | None = ...,
        control: global___ControlMessage | None = ...,
        senderRecord: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["control", b"control", "senderRecord", b"senderRecord"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["control", b"control", "publish", b"publish", "senderRecord", b"senderRecord", "subscriptions", b"subscriptions"]) -> None: ...

global___RPC = RPC

@typing.final
class Message(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FROM_ID_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    SEQNO_FIELD_NUMBER: builtins.int
    TOPICIDS_FIELD_NUMBER: builtins.int
    SIGNATURE_FIELD_NUMBER: builtins.int
    KEY_FIELD_NUMBER: builtins.int
    from_id: builtins.bytes
    data: builtins.bytes
    seqno: builtins.bytes
    signature: builtins.bytes
    key: builtins.bytes
    @property
    def topicIDs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        from_id: builtins.bytes | None = ...,
        data: builtins.bytes | None = ...,
        seqno: builtins.bytes | None = ...,
        topicIDs: collections.abc.Iterable[builtins.str] | None = ...,
        signature: builtins.bytes | None = ...,
        key: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["data", b"data", "from_id", b"from_id", "key", b"key", "seqno", b"seqno", "signature", b"signature"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["data", b"data", "from_id", b"from_id", "key", b"key", "seqno", b"seqno", "signature", b"signature", "topicIDs", b"topicIDs"]) -> None: ...

global___Message = Message

@typing.final
class ControlMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IHAVE_FIELD_NUMBER: builtins.int
    IWANT_FIELD_NUMBER: builtins.int
    GRAFT_FIELD_NUMBER: builtins.int
    PRUNE_FIELD_NUMBER: builtins.int
    IDONTWANT_FIELD_NUMBER: builtins.int
    @property
    def ihave(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ControlIHave]: ...
    @property
    def iwant(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ControlIWant]: ...
    @property
    def graft(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ControlGraft]: ...
    @property
    def prune(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ControlPrune]: ...
    @property
    def idontwant(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ControlIDontWant]: ...
    def __init__(
        self,
        *,
        ihave: collections.abc.Iterable[global___ControlIHave] | None = ...,
        iwant: collections.abc.Iterable[global___ControlIWant] | None = ...,
        graft: collections.abc.Iterable[global___ControlGraft] | None = ...,
        prune: collections.abc.Iterable[global___ControlPrune] | None = ...,
        idontwant: collections.abc.Iterable[global___ControlIDontWant] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["graft", b"graft", "idontwant", b"idontwant", "ihave", b"ihave", "iwant", b"iwant", "prune", b"prune"]) -> None: ...

global___ControlMessage = ControlMessage

@typing.final
class ControlIHave(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOPICID_FIELD_NUMBER: builtins.int
    MESSAGEIDS_FIELD_NUMBER: builtins.int
    topicID: builtins.str
    @property
    def messageIDs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        topicID: builtins.str | None = ...,
        messageIDs: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["topicID", b"topicID"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["messageIDs", b"messageIDs", "topicID", b"topicID"]) -> None: ...

global___ControlIHave = ControlIHave

@typing.final
class ControlIWant(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGEIDS_FIELD_NUMBER: builtins.int
    @property
    def messageIDs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        messageIDs: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["messageIDs", b"messageIDs"]) -> None: ...

global___ControlIWant = ControlIWant

@typing.final
class ControlGraft(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOPICID_FIELD_NUMBER: builtins.int
    topicID: builtins.str
    def __init__(
        self,
        *,
        topicID: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["topicID", b"topicID"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["topicID", b"topicID"]) -> None: ...

global___ControlGraft = ControlGraft

@typing.final
class ControlPrune(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOPICID_FIELD_NUMBER: builtins.int
    PEERS_FIELD_NUMBER: builtins.int
    BACKOFF_FIELD_NUMBER: builtins.int
    topicID: builtins.str
    backoff: builtins.int
    @property
    def peers(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___PeerInfo]: ...
    def __init__(
        self,
        *,
        topicID: builtins.str | None = ...,
        peers: collections.abc.Iterable[global___PeerInfo] | None = ...,
        backoff: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["backoff", b"backoff", "topicID", b"topicID"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["backoff", b"backoff", "peers", b"peers", "topicID", b"topicID"]) -> None: ...

global___ControlPrune = ControlPrune

@typing.final
class ControlIDontWant(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGEIDS_FIELD_NUMBER: builtins.int
    @property
    def messageIDs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(
        self,
        *,
        messageIDs: collections.abc.Iterable[builtins.bytes] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["messageIDs", b"messageIDs"]) -> None: ...

global___ControlIDontWant = ControlIDontWant

@typing.final
class PeerInfo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PEERID_FIELD_NUMBER: builtins.int
    SIGNEDPEERRECORD_FIELD_NUMBER: builtins.int
    peerID: builtins.bytes
    signedPeerRecord: builtins.bytes
    def __init__(
        self,
        *,
        peerID: builtins.bytes | None = ...,
        signedPeerRecord: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["peerID", b"peerID", "signedPeerRecord", b"signedPeerRecord"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["peerID", b"peerID", "signedPeerRecord", b"signedPeerRecord"]) -> None: ...

global___PeerInfo = PeerInfo

@typing.final
class TopicDescriptor(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class AuthOpts(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class _AuthMode:
            ValueType = typing.NewType("ValueType", builtins.int)
            V: typing_extensions.TypeAlias = ValueType

        class _AuthModeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[TopicDescriptor.AuthOpts._AuthMode.ValueType], builtins.type):
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
            NONE: TopicDescriptor.AuthOpts._AuthMode.ValueType  # 0
            """no authentication, anyone can publish"""
            KEY: TopicDescriptor.AuthOpts._AuthMode.ValueType  # 1
            """only messages signed by keys in the topic descriptor are accepted"""
            WOT: TopicDescriptor.AuthOpts._AuthMode.ValueType  # 2
            """web of trust, certificates can allow publisher set to grow"""

        class AuthMode(_AuthMode, metaclass=_AuthModeEnumTypeWrapper): ...
        NONE: TopicDescriptor.AuthOpts.AuthMode.ValueType  # 0
        """no authentication, anyone can publish"""
        KEY: TopicDescriptor.AuthOpts.AuthMode.ValueType  # 1
        """only messages signed by keys in the topic descriptor are accepted"""
        WOT: TopicDescriptor.AuthOpts.AuthMode.ValueType  # 2
        """web of trust, certificates can allow publisher set to grow"""

        MODE_FIELD_NUMBER: builtins.int
        KEYS_FIELD_NUMBER: builtins.int
        mode: global___TopicDescriptor.AuthOpts.AuthMode.ValueType
        @property
        def keys(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]:
            """root keys to trust"""

        def __init__(
            self,
            *,
            mode: global___TopicDescriptor.AuthOpts.AuthMode.ValueType | None = ...,
            keys: collections.abc.Iterable[builtins.bytes] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["mode", b"mode"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["keys", b"keys", "mode", b"mode"]) -> None: ...

    @typing.final
    class EncOpts(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        class _EncMode:
            ValueType = typing.NewType("ValueType", builtins.int)
            V: typing_extensions.TypeAlias = ValueType

        class _EncModeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[TopicDescriptor.EncOpts._EncMode.ValueType], builtins.type):
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
            NONE: TopicDescriptor.EncOpts._EncMode.ValueType  # 0
            """no encryption, anyone can read"""
            SHAREDKEY: TopicDescriptor.EncOpts._EncMode.ValueType  # 1
            """messages are encrypted with shared key"""
            WOT: TopicDescriptor.EncOpts._EncMode.ValueType  # 2
            """web of trust, certificates can allow publisher set to grow"""

        class EncMode(_EncMode, metaclass=_EncModeEnumTypeWrapper): ...
        NONE: TopicDescriptor.EncOpts.EncMode.ValueType  # 0
        """no encryption, anyone can read"""
        SHAREDKEY: TopicDescriptor.EncOpts.EncMode.ValueType  # 1
        """messages are encrypted with shared key"""
        WOT: TopicDescriptor.EncOpts.EncMode.ValueType  # 2
        """web of trust, certificates can allow publisher set to grow"""

        MODE_FIELD_NUMBER: builtins.int
        KEYHASHES_FIELD_NUMBER: builtins.int
        mode: global___TopicDescriptor.EncOpts.EncMode.ValueType
        @property
        def keyHashes(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]:
            """the hashes of the shared keys used (salted)"""

        def __init__(
            self,
            *,
            mode: global___TopicDescriptor.EncOpts.EncMode.ValueType | None = ...,
            keyHashes: collections.abc.Iterable[builtins.bytes] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["mode", b"mode"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["keyHashes", b"keyHashes", "mode", b"mode"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    AUTH_FIELD_NUMBER: builtins.int
    ENC_FIELD_NUMBER: builtins.int
    name: builtins.str
    @property
    def auth(self) -> global___TopicDescriptor.AuthOpts: ...
    @property
    def enc(self) -> global___TopicDescriptor.EncOpts: ...
    def __init__(
        self,
        *,
        name: builtins.str | None = ...,
        auth: global___TopicDescriptor.AuthOpts | None = ...,
        enc: global___TopicDescriptor.EncOpts | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["auth", b"auth", "enc", b"enc", "name", b"name"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["auth", b"auth", "enc", b"enc", "name", b"name"]) -> None: ...

global___TopicDescriptor = TopicDescriptor



import ipaddress
import logging

import miniupnpc  # type: ignore[import-untyped]
import trio

logger = logging.getLogger("nanolib.discovery.upnp")


class UpnpManager:
    """
    A simple, self-contained manager for UPnP port mapping that can be used
    alongside a libp2p Host.
    """

    def __init__(self) -> None:
        self._gateway = miniupnpc.UPnP()
        self._lan_addr: str | None = None
        self._external_ip: str | None = None

    def _validate_igd(self) -> bool:
        """
        Validate that the selected device is actually an Internet Gateway Device.

        :return: True if the device is a valid IGD, False otherwise.
        """
        try:
            # Check if we can get basic IGD information
            if not self._gateway.lanaddr:
                logger.debug("IGD validation failed: No LAN address available")
                return False

            # Try to get the external IP - this is a good test of IGD validity
            external_ip = self._gateway.externalipaddress()
            if not external_ip:
                logger.debug("IGD validation failed: No external IP address available")
                return False

            # Additional validation: check if we can get the connection type
            # This is a more advanced IGD feature that non-IGD devices typically
            # don't support
            try:
                connection_type = self._gateway.connectiontype()
                if connection_type:
                    logger.debug(f"IGD connection type: {connection_type}")
            except Exception:
                # Connection type is optional, so we don't fail if it's not available
                logger.debug("IGD connection type not available (this is optional)")

            return True
        except Exception as e:
            logger.debug(f"IGD validation failed: {e}")
            return False

    async def discover(self) -> bool:
        """
        Discover the UPnP IGD on the network.

        :return: True if a gateway is found, False otherwise.
        """
        logger.debug("Discovering UPnP gateway...")
        try:
            try:
                num_devices = await trio.to_thread.run_sync(self._gateway.discover)
            except Exception as e:
                # The miniupnpc library has a documented quirk where `discover()` can
                # raise an exception with the message "Success" on some platforms
                # (particularly Windows) due to inconsistent error handling in the C
                # library. This is a known issue in miniupnpc where successful
                # discovery sometimes raises an exception instead of returning a count.
                # See: https://github.com/miniupnp/miniupnp/issues/
                if str(e) == "Success":  # type: ignore
                    num_devices = 1
                else:
                    logger.exception("UPnP discovery exception")
                    return False

            if num_devices > 0:
                logger.debug(f"Found {num_devices} UPnP device(s), selecting IGD...")
                try:
                    await trio.to_thread.run_sync(self._gateway.selectigd)
                except Exception as e:
                    logger.error(f"Failed to select IGD: {e}")
                    logger.error(
                        "UPnP devices were found, but none are valid Internet "
                        "Gateway Devices. Check your router's UPnP/IGD settings."
                    )
                    return False

                # Validate that the selected device is actually an IGD
                if not await trio.to_thread.run_sync(self._validate_igd):
                    logger.error(
                        "Selected UPnP device is not a valid Internet Gateway Device. "
                        "The device may be a smart home device or other UPnP device "
                        "that doesn't support port mapping."
                    )
                    return False

                self._lan_addr = self._gateway.lanaddr
                self._external_ip = await trio.to_thread.run_sync(
                    self._gateway.externalipaddress
                )
                logger.debug(f"UPnP gateway found: {self._external_ip}")

                if self._external_ip is None:
                    logger.error("Gateway did not return an external IP address")
                    return False

                ip_obj = ipaddress.ip_address(self._external_ip)
                if ip_obj.is_private:
                    logger.warning(
                        "UPnP gateway has a private IP; you may be behind a double NAT."
                    )
                    return False
                return True
            else:
                logger.debug("No UPnP devices found")
                return False
        except Exception:
            logger.exception("UPnP discovery failed")
            return False

    async def add_port_mapping(self, port: int, protocol: str = "TCP") -> bool:
        """
        Request a new port mapping from the gateway.

        :param port: the internal port to map
        :param protocol: the protocol to map (TCP or UDP)
        :return: True on success, False otherwise
        """
        try:
            port = int(port)
            if not 0 < port < 65536:
                logger.error(f"Invalid port number for mapping: {port}")
                return False
        except (ValueError, TypeError):
            logger.error(f"Invalid port value: {port}")
            return False
        if port < 1024:
            logger.warning(
                f"Mapping a well-known (privileged) port ({port}) may fail or "
                "require root."
            )

        if not self._lan_addr:
            logger.error(
                "Cannot add port mapping: discovery has not been run successfully."
            )
            return False

        logger.debug(f"Requesting UPnP mapping for {protocol} port {port}...")
        try:
            await trio.to_thread.run_sync(
                lambda: self._gateway.addportmapping(
                    port, protocol, self._lan_addr, port, "py-libp2p", ""
                )
            )
            logger.info(
                f"Successfully mapped external port {self._external_ip}:{port} "
                f"to internal port {self._lan_addr}:{port}"
            )
            return True
        except Exception:
            logger.exception(f"Failed to map port {port}")
        return False

    async def remove_port_mapping(self, port: int, protocol: str = "TCP") -> bool:
        """
        Remove an existing port mapping.

        :param port: the external port to unmap
        :param protocol: the protocol (TCP or UDP)
        :return: True on success, False otherwise
        """
        try:
            port = int(port)
            if not 0 < port < 65536:
                logger.error(f"Invalid port number for removal: {port}")
                return False
        except (ValueError, TypeError):
            logger.error(f"Invalid port value: {port}")
            return False

        logger.debug(f"Removing UPnP mapping for {protocol} port {port}...")
        try:
            await trio.to_thread.run_sync(
                lambda: self._gateway.deleteportmapping(port, protocol)
            )
            logger.info(f"Successfully removed mapping for port {port}")
            return True
        except Exception:
            logger.exception(f"Failed to remove mapping for port {port}")
        return False

    def get_external_ip(self) -> str | None:
        return self._external_ip



from nanolib.io.abc import (
    Reader,
)
from nanolib.io.exceptions import (
    IncompleteReadError,
)

DEFAULT_RETRY_READ_COUNT = 100


async def read_exactly(
    reader: Reader, n: int, retry_count: int = DEFAULT_RETRY_READ_COUNT
) -> bytes:
    """
    Read exactly n bytes from the reader.

    This function attempts to read exactly n bytes from the reader, retrying
    up to retry_count times if partial data is received. If the connection
    closes before n bytes are received, raises IncompleteReadError with
    enhanced error messages that include transport context (if available).

    The error message includes:

    - The expected and actual number of bytes received
    - Transport type and connection duration (if reader provides conn_state())
    - Context about possible causes (peer closure, network issues,
      transport-specific problems)

    Args:
        reader: The reader to read from
        n: Number of bytes to read
        retry_count: Maximum number of retries if partial data is received

    Returns:
        bytes: Exactly n bytes of data

    Raises:
        IncompleteReadError: If the connection closes before n bytes are received.
            The error message includes transport context if available via
            reader.conn_state().

    .. note::
        Relies on exceptions to break out on erroneous conditions, like EOF.

    """
    buffer = bytearray()
    buffer.extend(await reader.read(n))

    for _ in range(retry_count):
        if len(buffer) < n:
            remaining = n - len(buffer)
            buffer.extend(await reader.read(remaining))

        else:
            return bytes(buffer)

    # Get transport context if available
    context_info = ""
    try:
        if hasattr(reader, "conn_state"):
            conn_state_method = getattr(reader, "conn_state")
            if callable(conn_state_method):
                state = conn_state_method()
                if isinstance(state, dict):
                    transport = state.get("transport", "unknown")
                    duration = state.get("connection_duration", 0)
                    context_info = (
                        f" (transport: {transport}, duration: {duration:.2f}s)"
                    )
    except Exception:
        # Gracefully handle if conn_state() is not available or fails
        pass

    raise IncompleteReadError(
        f"Connection closed during read operation: expected {n} bytes but "
        f"received {len(buffer)} bytes{context_info}"
    )



"""
monitoring and alerting system.

This module provides comprehensive monitoring capabilities including
OpenMetrics format export, real-time metrics collection, connection
duration tracking, and protocol-specific metrics. Now integrated with
Prometheus metrics export for compatibility with go-libp2p dashboards.
"""

from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field
from enum import Enum
import threading
import time
from typing import Any

from .prometheus_exporter import PrometheusExporter, create_prometheus_exporter


class MetricType(Enum):
    """Types of metrics for monitoring."""

    COUNTER = "counter"
    GAUGE = "gauge"
    HISTOGRAM = "histogram"
    SUMMARY = "summary"


class AlertSeverity(Enum):
    """Alert severity levels."""

    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    CRITICAL = "critical"


@dataclass
class Metric:
    """A single metric measurement."""

    name: str
    value: int | float
    metric_type: MetricType
    labels: dict[str, str] = field(default_factory=dict)
    timestamp: float = field(default_factory=time.time)
    help_text: str = ""


@dataclass
class Alert:
    """An alert condition."""

    name: str
    condition: str
    severity: AlertSeverity
    message: str
    threshold: int | float
    current_value: int | float
    timestamp: float = field(default_factory=time.time)
    resolved: bool = False


@dataclass
class ConnectionDuration:
    """Connection duration tracking."""

    connection_id: str
    peer_id: str
    start_time: float
    end_time: float | None = None
    duration: float | None = None
    protocol: str = "unknown"
    direction: str = "unknown"


class Monitor:
    """
    Monitoring and alerting system.

    Provides monitoring capabilities such as real-time metrics collection.
    """

    def __init__(
        self,
        buffer_size: int = 1000,
        alert_thresholds: dict[str, float] | None = None,
        enable_connection_tracking: bool = True,
        enable_protocol_metrics: bool = True,
        enable_prometheus: bool = False,
        prometheus_port: int = 8000,
        prometheus_exporter: PrometheusExporter | None = None,
    ) -> None:
        """
        Initialize production monitor.

        Args:
            buffer_size: Maximum number of metrics to buffer
            alert_thresholds: Custom alert thresholds
            enable_connection_tracking: Enable connection duration tracking
            enable_protocol_metrics: Enable protocol-specific metrics
            enable_prometheus: Enable Prometheus metrics export
            prometheus_port: Port for Prometheus metrics server
            prometheus_exporter: Custom Prometheus exporter instance

        """
        self.buffer_size = buffer_size
        self.alert_thresholds = alert_thresholds or {}
        self.enable_connection_tracking = enable_connection_tracking
        self.enable_protocol_metrics = enable_protocol_metrics

        # Prometheus integration
        self.prometheus_exporter: PrometheusExporter | None
        if prometheus_exporter is not None:
            self.prometheus_exporter = prometheus_exporter
        elif enable_prometheus:
            self.prometheus_exporter = create_prometheus_exporter(
                port=prometheus_port, enable_server=True
            )
        else:
            self.prometheus_exporter = None

        # Metrics storage
        self.metrics_buffer: deque[Metric] = deque(maxlen=buffer_size)
        self.alerts: list[Alert] = []
        self._lock = threading.RLock()

        # Connection tracking
        self._connection_durations: dict[str, ConnectionDuration] = {}
        self._connection_establishment_times: dict[str, float] = {}

        # Protocol metrics
        self._protocol_metrics: dict[str, dict[str, Any]] = {}

        # System metrics
        self._start_time = time.time()
        self._total_metrics_recorded = 0
        self._total_alerts_triggered = 0

        # Default alert thresholds
        self._default_thresholds = {
            "connection_limit_percentage": 80.0,
            "memory_limit_percentage": 80.0,
            "stream_limit_percentage": 80.0,
            "error_rate_percentage": 5.0,
            "response_time_ms": 1000.0,
        }

        # Merge custom thresholds
        self.alert_thresholds = {**self._default_thresholds, **self.alert_thresholds}

    def _export_to_prometheus(self, metric: Metric) -> None:
        """
        Export metric to Prometheus if exporter is available.

        Args:
            metric: The metric to export

        """
        if not self.prometheus_exporter:
            return

        try:
            # Map Monitor metrics to Prometheus format
            if metric.name.startswith("libp2p_connections_"):
                metric.labels.get("direction", "unknown")
                if metric.name.endswith("_total"):
                    # This is a connection establishment/closure event
                    # We'll let the resource manager handle direct Prometheus updates
                    pass

            elif metric.name.startswith("libp2p_resource_usage"):
                resource_type = metric.labels.get("resource_type", "unknown")
                if metric.name.endswith("_percentage"):
                    # Update blocked resources if percentage is high
                    if metric.value >= 90:  # High usage threshold
                        self.prometheus_exporter.record_blocked_resource(
                            direction="", scope="system", resource=resource_type
                        )

            elif metric.name == "libp2p_errors_total":
                metric.labels.get("error_type", "unknown")
                # Could record error-based blocked resources

        except Exception as e:
            # Don't let Prometheus export failures affect main monitoring
            print(f"Warning: Failed to export metric to Prometheus: {e}")

    def record_metric(self, metric: Metric) -> None:
        """
        Record a metric with buffering for performance.

        Args:
            metric: The metric to record

        """
        with self._lock:
            self.metrics_buffer.append(metric)
            self._total_metrics_recorded += 1
            self._check_alerts(metric)

            # Export to Prometheus if enabled
            self._export_to_prometheus(metric)

    def record_connection_establishment(
        self,
        connection_id: str,
        peer_id: str,
        protocol: str = "unknown",
        direction: str = "unknown",
    ) -> None:
        """
        Record connection establishment for duration tracking.

        Args:
            connection_id: Unique connection identifier
            peer_id: Peer identifier
            protocol: Protocol used for connection
            direction: Connection direction (inbound/outbound)

        """
        if not self.enable_connection_tracking:
            return

        with self._lock:
            current_time = time.time()
            self._connection_establishment_times[connection_id] = current_time

            # Create connection duration tracker
            duration_tracker = ConnectionDuration(
                connection_id=connection_id,
                peer_id=peer_id,
                start_time=current_time,
                protocol=protocol,
                direction=direction,
            )
            self._connection_durations[connection_id] = duration_tracker

            # Record establishment metric
            establishment_metric = Metric(
                name="libp2p_connections_established_total",
                value=1,
                metric_type=MetricType.COUNTER,
                labels={
                    "protocol": protocol,
                    "direction": direction,
                    "peer_id": peer_id,
                },
                help_text="Total number of connections established",
            )
            self.record_metric(establishment_metric)

    def record_connection_closed(
        self, connection_id: str, reason: str = "unknown"
    ) -> None:
        """
        Record connection closure and calculate duration.

        Args:
            connection_id: Unique connection identifier
            reason: Reason for connection closure

        """
        if not self.enable_connection_tracking:
            return

        with self._lock:
            if connection_id not in self._connection_durations:
                return

            current_time = time.time()
            duration_tracker = self._connection_durations[connection_id]
            duration_tracker.end_time = current_time
            duration_tracker.duration = current_time - duration_tracker.start_time

            # Record connection duration metric
            duration_metric = Metric(
                name="libp2p_connection_duration_seconds",
                value=duration_tracker.duration,
                metric_type=MetricType.HISTOGRAM,
                labels={
                    "protocol": duration_tracker.protocol,
                    "direction": duration_tracker.direction,
                    "reason": reason,
                },
                help_text="Connection duration in seconds",
            )
            self.record_metric(duration_metric)

            # Record connection closed metric
            closed_metric = Metric(
                name="libp2p_connections_closed_total",
                value=1,
                metric_type=MetricType.COUNTER,
                labels={
                    "protocol": duration_tracker.protocol,
                    "direction": duration_tracker.direction,
                    "reason": reason,
                },
                help_text="Total number of connections closed",
            )
            self.record_metric(closed_metric)

            # Clean up
            del self._connection_durations[connection_id]
            if connection_id in self._connection_establishment_times:
                del self._connection_establishment_times[connection_id]

    def record_protocol_metric(
        self,
        protocol: str,
        metric_name: str,
        value: int | float,
        labels: dict[str, str] | None = None,
    ) -> None:
        """
        Record protocol-specific metric.

        Args:
            protocol: Protocol name
            metric_name: Name of the metric
            value: Metric value
            labels: Additional labels

        """
        if not self.enable_protocol_metrics:
            return

        with self._lock:
            # Update protocol metrics
            if protocol not in self._protocol_metrics:
                self._protocol_metrics[protocol] = {}
            protocol_metrics: dict[str, Any] = self._protocol_metrics[protocol]
            protocol_metrics[metric_name] = value

            # Create metric
            metric_labels = labels or {}
            metric_labels["protocol"] = protocol

            metric = Metric(
                name=f"libp2p_protocol_{metric_name}",
                value=value,
                metric_type=MetricType.GAUGE,
                labels=metric_labels,
                help_text=f"Protocol-specific metric: {metric_name}",
            )
            self.record_metric(metric)

    def record_resource_usage(
        self,
        resource_type: str,
        current_value: int | float,
        limit_value: int | float,
        labels: dict[str, str] | None = None,
    ) -> None:
        """
        Record resource usage metrics.

        Args:
            resource_type: Type of resource (connections, memory, streams)
            current_value: Current usage value
            limit_value: Maximum limit value
            labels: Additional labels

        """
        metric_labels = labels or {}
        metric_labels["resource_type"] = resource_type

        # Record current usage
        current_metric = Metric(
            name="libp2p_resource_usage_current",
            value=current_value,
            metric_type=MetricType.GAUGE,
            labels=metric_labels,
            help_text=f"Current {resource_type} usage",
        )
        self.record_metric(current_metric)

        # Record limit
        limit_metric = Metric(
            name="libp2p_resource_usage_limit",
            value=limit_value,
            metric_type=MetricType.GAUGE,
            labels=metric_labels,
            help_text=f"Maximum {resource_type} limit",
        )
        self.record_metric(limit_metric)

        # Record percentage
        percentage = (current_value / limit_value * 100) if limit_value > 0 else 0
        percentage_metric = Metric(
            name="libp2p_resource_usage_percentage",
            value=percentage,
            metric_type=MetricType.GAUGE,
            labels=metric_labels,
            help_text=f"{resource_type} usage percentage",
        )
        self.record_metric(percentage_metric)

    def record_error(
        self, error_type: str, error_code: str, labels: dict[str, str] | None = None
    ) -> None:
        """
        Record error metrics.

        Args:
            error_type: Type of error
            error_code: Error code
            labels: Additional labels

        """
        metric_labels = labels or {}
        metric_labels["error_type"] = error_type
        metric_labels["error_code"] = error_code

        error_metric = Metric(
            name="libp2p_errors_total",
            value=1,
            metric_type=MetricType.COUNTER,
            labels=metric_labels,
            help_text="Total number of errors by type and code",
        )
        self.record_metric(error_metric)

    def record_blocked_resource(
        self,
        resource_type: str,
        direction: str = "unknown",
        scope: str = "system",
        count: int = 1,
    ) -> None:
        """
        Record blocked resource event for both internal monitoring and Prometheus.

        Args:
            resource_type: Type of resource (connection, stream, memory)
            direction: Direction (inbound, outbound, or empty)
            scope: Resource scope (system, transient, peer, etc.)
            count: Number of resources blocked

        """
        # Record internal metric
        metric = Metric(
            name="libp2p_rcmgr_blocked_resources",
            value=count,
            metric_type=MetricType.COUNTER,
            labels={
                "resource": resource_type,
                "direction": direction,
                "scope": scope,
            },
            help_text="Number of blocked resources",
        )
        self.record_metric(metric)

        # Export directly to Prometheus
        if self.prometheus_exporter:
            self.prometheus_exporter.record_blocked_resource(
                direction=direction,
                scope=scope,
                resource=resource_type,
                count=count,
            )

    def record_peer_resource_change(
        self,
        peer_id: str,
        resource_type: str,
        direction: str,
        old_value: int,
        new_value: int,
    ) -> None:
        """
        Record peer-level resource changes for histogram metrics.

        Args:
            peer_id: Peer identifier
            resource_type: Type of resource (connections, streams, memory)
            direction: Direction for connections/streams
            old_value: Previous resource count/amount
            new_value: New resource count/amount

        """
        if not self.prometheus_exporter:
            return

        try:
            if resource_type == "connections":
                self.prometheus_exporter.record_peer_connections(
                    peer_id=peer_id,
                    direction=direction,
                    old_count=old_value,
                    new_count=new_value,
                )
            elif resource_type == "streams":
                self.prometheus_exporter.record_peer_streams(
                    peer_id=peer_id,
                    direction=direction,
                    old_count=old_value,
                    new_count=new_value,
                )
            elif resource_type == "memory":
                self.prometheus_exporter.record_peer_memory(
                    peer_id=peer_id,
                    old_bytes=old_value,
                    new_bytes=new_value,
                )
        except Exception as e:
            print(f"Warning: Failed to record peer resource change: {e}")

    def get_prometheus_metrics(self) -> str | None:
        """
        Get metrics in Prometheus text format.

        Returns:
            Metrics in Prometheus format or None if exporter not available

        """
        if self.prometheus_exporter:
            return self.prometheus_exporter.get_metrics_text()
        return None

    def _check_alerts(self, metric: Metric) -> None:
        """
        Check for alert conditions based on metric.

        Args:
            metric: The metric to check for alerts

        """
        # Check resource usage alerts
        if metric.name == "libp2p_resource_usage_percentage":
            resource_type = metric.labels.get("resource_type", "unknown")
            threshold_key = f"{resource_type}_limit_percentage"
            threshold = self.alert_thresholds.get(threshold_key, 80.0)

            if metric.value >= threshold:
                alert = Alert(
                    name=f"{resource_type}_limit_high",
                    condition=f"{resource_type} usage >= {threshold}%",
                    severity=(
                        AlertSeverity.WARNING
                        if metric.value < 95
                        else AlertSeverity.CRITICAL
                    ),
                    message=(
                        f"{resource_type} usage is {metric.value:.1f}% "
                        f"(threshold: {threshold}%)"
                    ),
                    threshold=threshold,
                    current_value=metric.value,
                )
                self._trigger_alert(alert)

        # Check error rate alerts
        elif metric.name == "libp2p_errors_total":
            # Calculate error rate (simplified)
            # error_rate_threshold = self.alert_thresholds.get(
            #     "error_rate_percentage", 5.0
            # )
            # This would need more sophisticated error rate calculation
            pass

    def _trigger_alert(self, alert: Alert) -> None:
        """
        Trigger an alert.

        Args:
            alert: The alert to trigger

        """
        with self._lock:
            self.alerts.append(alert)
            self._total_alerts_triggered += 1

    def export_openmetrics(self) -> str:
        """
        Export metrics in OpenMetrics format.

        Returns:
            str: Metrics in OpenMetrics format

        """
        with self._lock:
            lines = []

            # Add header
            lines.append(
                "# HELP libp2p_monitor_uptime_seconds Monitor uptime in seconds"
            )
            lines.append("# TYPE libp2p_monitor_uptime_seconds gauge")
            lines.append(
                f"libp2p_monitor_uptime_seconds {time.time() - self._start_time}"
            )
            lines.append("")

            # Add metrics
            for metric in self.metrics_buffer:
                # Add help text
                if metric.help_text:
                    lines.append(f"# HELP {metric.name} {metric.help_text}")

                # Add type
                lines.append(f"# TYPE {metric.name} {metric.metric_type.value}")

                # Add metric with labels
                if metric.labels:
                    label_pairs = [f'{k}="{v}"' for k, v in metric.labels.items()]
                    labels_str = "{" + ",".join(label_pairs) + "}"
                    lines.append(
                        f"{metric.name}{labels_str} {metric.value} "
                        f"{int(metric.timestamp * 1000)}"
                    )
                else:
                    lines.append(
                        f"{metric.name} {metric.value} {int(metric.timestamp * 1000)}"
                    )
                lines.append("")

            return "\n".join(lines)

    def get_metrics_summary(self) -> dict[str, Any]:
        """
        Get a summary of all metrics.

        Returns:
            Dict containing metrics summary

        """
        with self._lock:
            return {
                "uptime_seconds": time.time() - self._start_time,
                "total_metrics_recorded": self._total_metrics_recorded,
                "total_alerts_triggered": self._total_alerts_triggered,
                "active_connections": len(self._connection_durations),
                "protocols_tracked": list(self._protocol_metrics.keys()),
                "recent_alerts": [
                    {
                        "name": alert.name,
                        "severity": alert.severity.value,
                        "message": alert.message,
                        "timestamp": alert.timestamp,
                        "resolved": alert.resolved,
                    }
                    for alert in self.alerts[-10:]  # Last 10 alerts
                ],
            }

    def get_connection_durations(self) -> list[dict[str, Any]]:
        """
        Get connection duration statistics.

        Returns:
            List of connection duration information

        """
        with self._lock:
            durations = []
            for conn_id, duration_tracker in self._connection_durations.items():
                durations.append(
                    {
                        "connection_id": conn_id,
                        "peer_id": duration_tracker.peer_id,
                        "protocol": duration_tracker.protocol,
                        "direction": duration_tracker.direction,
                        "start_time": duration_tracker.start_time,
                        "end_time": duration_tracker.end_time,
                        "duration": duration_tracker.duration,
                        "is_active": duration_tracker.end_time is None,
                    }
                )
            return durations

    def get_protocol_metrics(self) -> dict[str, dict[str, Any]]:
        """
        Get protocol-specific metrics.

        Returns:
            Dict of protocol metrics

        """
        with self._lock:
            return dict(self._protocol_metrics)

    def clear_old_metrics(self, max_age_seconds: float = 3600) -> None:
        """
        Clear old metrics to prevent memory buildup.

        Args:
            max_age_seconds: Maximum age of metrics to keep

        """
        with self._lock:
            current_time = time.time()
            cutoff_time = current_time - max_age_seconds

            # Clear old metrics from buffer
            while (
                self.metrics_buffer and self.metrics_buffer[0].timestamp < cutoff_time
            ):
                self.metrics_buffer.popleft()

            # Clear old alerts
            self.alerts = [
                alert for alert in self.alerts if alert.timestamp >= cutoff_time
            ]

    def reset(self) -> None:
        """Reset all monitoring data."""
        with self._lock:
            self.metrics_buffer.clear()
            self.alerts.clear()
            self._connection_durations.clear()
            self._connection_establishment_times.clear()
            self._protocol_metrics.clear()
            self._total_metrics_recorded = 0
            self._total_alerts_triggered = 0
            self._start_time = time.time()

            # Reset Prometheus metrics if available
            if self.prometheus_exporter:
                self.prometheus_exporter.reset()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: libp2p/bitswap/pb/dag_pb.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1elibp2p/bitswap/pb/dag_pb.proto\x12\x06\x64\x61g_pb\"3\n\x06PBLink\x12\x0c\n\x04Hash\x18\x01 \x01(\x0c\x12\x0c\n\x04Name\x18\x02 \x01(\t\x12\r\n\x05Tsize\x18\x03 \x01(\x04\"5\n\x06PBNode\x12\x1d\n\x05Links\x18\x02 \x03(\x0b\x32\x0e.dag_pb.PBLink\x12\x0c\n\x04\x44\x61ta\x18\x01 \x01(\x0c')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nanolib.bitswap.pb.dag_pb_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_PBLINK']._serialized_start=42
  _globals['_PBLINK']._serialized_end=93
  _globals['_PBNODE']._serialized_start=95
  _globals['_PBNODE']._serialized_end=148
# @@protoc_insertion_point(module_scope)



import logging

from nanolib.records.utils import InvalidRecordType, split_key

logger = logging.getLogger("records-validator")


class ErrBetterRecord(Exception):
    def __init__(self, key: str, value: bytes):
        self.key = key
        self.value = value
        super().__init__(f'Found better value for "{key}')


class Validator:
    """Base class for all validators"""

    def validate(self, key: str, value: bytes) -> None:
        raise NotImplementedError

    def select(self, key: str, values: list[bytes]) -> int:
        raise NotImplementedError


class NamespacedValidator:
    """
    Manages a collection of validators, each associated with a specific namespace.
    """

    def __init__(self, validators: dict[str, Validator]):
        self._validators = validators

    def validator_by_key(self, key: str) -> Validator | None:
        """
        Retrieve the validator responsible for the given key's namespace.

        Args:
            key (str): A namespaced key in the form "namespace/value".

        Returns:
            Optional[Validator]: The matching validator, or None if not found.

        """
        try:
            ns, _ = split_key(key)
        except InvalidRecordType:
            return None
        return self._validators.get(ns)

    def validate(self, key: str, value: bytes) -> None:
        """
        Validate a key-value pair using the appropriate namespaced validator.

        Args:
            key (str): The namespaced key (e.g., "pk/Qm...").
            value (bytes): The value to be validated.

        Raises:
            InvalidRecordType: If no matching validator is found.
            Exception: Propagates any exception raised by the sub-validator.

        """
        validator = self.validator_by_key(key)
        if validator is None:
            raise InvalidRecordType("Invalid record keytype")
        return validator.validate(key, value)

    def select(self, key: str, values: list[bytes]) -> int:
        """
        Choose the best value from a list using the namespaced validator.

        Args:
            key (str): The namespaced key used to find the validator.
            values (List[bytes]): List of candidate values to choose from.

        Returns:
            int: Index of the selected best value in the input list.

        Raises:
            ValueError: If the values list is empty.
            InvalidRecordType: If no matching validator is found.

        """
        if not values:
            raise ValueError("Can't select from empty value list")
        validator = self.validator_by_key(key)
        if validator is None:
            raise InvalidRecordType("Invalid record keytype")
        return validator.select(key, values)

    def add_validator(self, namespace: str, validator: Validator) -> None:
        """
        Add or update a validator for a specific namespace.

        Args:
            namespace (str): The namespace string (e.g., "pk", "myapp").
            validator (Validator): A Validator instance to handle validation
            for this namespace.

        """
        self._validators[namespace] = validator



"""
Synchronous datastore interfaces for persistent peer storage.

This module defines the synchronous interfaces that all sync datastore
implementations must follow, providing the core operations needed for
persistent peer storage without async/await.
"""

from abc import ABC, abstractmethod
from collections.abc import Iterator


class IDatastoreSync(ABC):
    """
    Abstract interface for synchronous datastore operations.

    This interface provides synchronous equivalents of the async datastore
    operations, allowing for pure synchronous peerstore implementations.
    """

    @abstractmethod
    def get(self, key: bytes) -> bytes | None:
        """
        Retrieve a value by key.

        Args:
            key: The key to look up

        Returns:
            The value if found, None otherwise

        Raises:
            KeyError: If the key is not found

        """
        pass

    @abstractmethod
    def put(self, key: bytes, value: bytes) -> None:
        """
        Store a key-value pair.

        Args:
            key: The key to store
            value: The value to store

        """
        pass

    @abstractmethod
    def delete(self, key: bytes) -> None:
        """
        Delete a key-value pair.

        Args:
            key: The key to delete

        """
        pass

    @abstractmethod
    def has(self, key: bytes) -> bool:
        """
        Check if a key exists.

        Args:
            key: The key to check

        Returns:
            True if the key exists, False otherwise

        """
        pass

    @abstractmethod
    def query(self, prefix: bytes) -> Iterator[tuple[bytes, bytes]]:
        """
        Query for keys with a given prefix.

        Args:
            prefix: The key prefix to search for

        Yields:
            Tuples of (key, value) for matching entries

        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Close the datastore and clean up resources.
        """
        pass

    @abstractmethod
    def sync(self, prefix: bytes) -> None:
        """
        Sync any pending writes to disk.

        Args:
            prefix: The key prefix to sync (empty bytes for all keys)

        """
        pass


class IBatchingDatastoreSync(IDatastoreSync):
    """
    Extended synchronous datastore interface that supports batched operations.

    This is useful for performance optimization when making multiple
    write operations.
    """

    @abstractmethod
    def batch(self) -> "IBatchSync":
        """
        Create a new batch for batched operations.

        Returns:
            A batch object for performing multiple operations atomically

        """
        pass


class IBatchSync(ABC):
    """
    Interface for synchronous batched datastore operations.
    """

    @abstractmethod
    def put(self, key: bytes, value: bytes) -> None:
        """
        Add a put operation to the batch.

        Args:
            key: The key to store
            value: The value to store

        """
        pass

    @abstractmethod
    def delete(self, key: bytes) -> None:
        """
        Add a delete operation to the batch.

        Args:
            key: The key to delete

        """
        pass

    @abstractmethod
    def commit(self) -> None:
        """
        Commit all operations in the batch atomically.
        """
        pass



"""
WebSocket multiaddr parsing utilities.
"""

from typing import NamedTuple

from multiaddr import Multiaddr
from multiaddr.protocols import Protocol


class ParsedWebSocketMultiaddr(NamedTuple):
    """Parsed WebSocket multiaddr information."""

    is_wss: bool
    sni: str | None
    rest_multiaddr: Multiaddr


def parse_websocket_multiaddr(maddr: Multiaddr) -> ParsedWebSocketMultiaddr:
    """
    Parse a WebSocket multiaddr and extract security information.

    :param maddr: The multiaddr to parse
    :return: Parsed WebSocket multiaddr information
    :raises ValueError: If the multiaddr is not a valid WebSocket multiaddr
    """
    # First validate that this is a valid WebSocket multiaddr
    if not is_valid_websocket_multiaddr(maddr):
        raise ValueError(f"Not a valid WebSocket multiaddr: {maddr}")

    protocols = list(maddr.protocols())

    # Find the WebSocket protocol and check for security
    is_wss = False
    sni = None
    ws_index = -1
    tls_index = -1
    sni_index = -1

    # Find protocol indices
    for i, protocol in enumerate(protocols):
        if protocol.name == "ws":
            ws_index = i
        elif protocol.name == "wss":
            ws_index = i
            is_wss = True
        elif protocol.name == "tls":
            tls_index = i
        elif protocol.name == "sni":
            sni_index = i
            sni = protocol.value

    if ws_index == -1:
        raise ValueError("Not a WebSocket multiaddr")

    # Handle /wss protocol (convert to /tls/ws internally)
    if is_wss and tls_index == -1:
        # Convert /wss to /tls/ws format
        # Remove /wss to get the base multiaddr
        without_wss = maddr.decapsulate(Multiaddr("/wss"))
        return ParsedWebSocketMultiaddr(
            is_wss=True, sni=None, rest_multiaddr=without_wss
        )

    # Handle /tls/ws and /tls/sni/.../ws formats
    if tls_index != -1:
        is_wss = True
        # Extract the base multiaddr (everything before /tls)
        # For /ip4/127.0.0.1/tcp/8080/tls/ws, we want /ip4/127.0.0.1/tcp/8080
        # Use multiaddr methods to properly extract the base
        rest_multiaddr = maddr
        # Remove /tls/ws or /tls/sni/.../ws from the end
        if sni_index != -1:
            # /tls/sni/example.com/ws format
            rest_multiaddr = rest_multiaddr.decapsulate(Multiaddr("/ws"))
            rest_multiaddr = rest_multiaddr.decapsulate(Multiaddr(f"/sni/{sni}"))
            rest_multiaddr = rest_multiaddr.decapsulate(Multiaddr("/tls"))
        else:
            # /tls/ws format
            rest_multiaddr = rest_multiaddr.decapsulate(Multiaddr("/ws"))
            rest_multiaddr = rest_multiaddr.decapsulate(Multiaddr("/tls"))
        return ParsedWebSocketMultiaddr(
            is_wss=is_wss, sni=sni, rest_multiaddr=rest_multiaddr
        )

    # Regular /ws multiaddr - remove /ws and any additional protocols
    rest_multiaddr = maddr.decapsulate(Multiaddr("/ws"))
    return ParsedWebSocketMultiaddr(
        is_wss=False, sni=None, rest_multiaddr=rest_multiaddr
    )


def is_valid_websocket_multiaddr(maddr: Multiaddr) -> bool:
    """
    Validate that a multiaddr has a valid WebSocket structure.

    :param maddr: The multiaddr to validate
    :return: True if valid WebSocket structure, False otherwise
    """
    try:
        # WebSocket multiaddr should have structure like:
        # /ip4/127.0.0.1/tcp/8080/ws (insecure)
        # /ip4/127.0.0.1/tcp/8080/wss (secure)
        # /ip4/127.0.0.1/tcp/8080/tls/ws (secure with TLS)
        # /ip4/127.0.0.1/tcp/8080/tls/sni/example.com/ws (secure with SNI)
        protocols: list[Protocol] = list(maddr.protocols())

        # Must have at least 3 protocols: network (ip4/ip6/dns4/dns6) + tcp + ws/wss
        if len(protocols) < 3:
            return False

        # First protocol should be a network protocol (ip4, ip6, dns, dns4, dns6)
        if protocols[0].name not in ["ip4", "ip6", "dns", "dns4", "dns6"]:
            return False

        # Second protocol should be tcp
        if protocols[1].name != "tcp":
            return False

        # Check for valid WebSocket protocols
        ws_protocols = ["ws", "wss"]
        tls_protocols = ["tls"]
        sni_protocols = ["sni"]

        # Find the WebSocket protocol
        ws_protocol_found = False
        tls_found = False
        # sni_found = False  # Not used currently

        for i, protocol in enumerate(protocols[2:], start=2):
            if protocol.name in ws_protocols:
                ws_protocol_found = True
                break
            elif protocol.name in tls_protocols:
                tls_found = True
            elif protocol.name in sni_protocols:
                pass  # sni_found = True  # Not used in current implementation

        if not ws_protocol_found:
            return False

        # Validate protocol sequence
        # For /ws: network + tcp + ws
        # For /wss: network + tcp + wss
        # For /tls/ws: network + tcp + tls + ws
        # For /tls/sni/example.com/ws: network + tcp + tls + sni + ws

        # Check if it's a simple /ws or /wss
        if len(protocols) == 3:
            return protocols[2].name in ["ws", "wss"]

        # Check for /tls/ws or /tls/sni/.../ws patterns
        if tls_found:
            # Must end with /ws (not /wss when using /tls)
            if protocols[-1].name != "ws":
                return False

            # Check for valid TLS sequence
            tls_index = None
            for i, protocol in enumerate(protocols[2:], start=2):
                if protocol.name == "tls":
                    tls_index = i
                    break

            if tls_index is None:
                return False

            # After tls, we can have sni, then ws
            remaining_protocols = protocols[tls_index + 1 :]
            if len(remaining_protocols) == 1:
                # /tls/ws
                return remaining_protocols[0].name == "ws"
            elif len(remaining_protocols) == 2:
                # /tls/sni/example.com/ws
                return (
                    remaining_protocols[0].name == "sni"
                    and remaining_protocols[1].name == "ws"
                )
            else:
                return False

        # If we have more than 3 protocols but no TLS, check for valid continuations
        # Allow additional protocols after the WebSocket protocol (like /p2p or /ipfs)
        valid_continuations = ["p2p", "ipfs"]

        # Find the WebSocket protocol index
        ws_index = None
        for i, protocol in enumerate(protocols):
            if protocol.name in ["ws", "wss"]:
                ws_index = i
                break

        if ws_index is not None:
            # Check protocols after the WebSocket protocol
            for i in range(ws_index + 1, len(protocols)):
                if protocols[i].name not in valid_continuations:
                    print(f"DEBUG: Invalid continuation protocol: {protocols[i].name}")
                    return False

        return True

    except Exception as e:
        print(f"DEBUG: Validation exception: {e}")
        return False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: libp2p/peer/pb/peer_record.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n libp2p/peer/pb/peer_record.proto\x12\x07peer.pb\"\x80\x01\n\nPeerRecord\x12\x0f\n\x07peer_id\x18\x01 \x01(\x0c\x12\x0b\n\x03seq\x18\x02 \x01(\x04\x12\x32\n\taddresses\x18\x03 \x03(\x0b\x32\x1f.peer.pb.PeerRecord.AddressInfo\x1a \n\x0b\x41\x64\x64ressInfo\x12\x11\n\tmultiaddr\x18\x01 \x01(\x0c\x42*Z(github.com/libp2p/go-libp2p/core/peer/pbb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nanolib.peer.pb.peer_record_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z(github.com/libp2p/go-libp2p/core/peer/pb'
  _globals['_PEERRECORD']._serialized_start=46
  _globals['_PEERRECORD']._serialized_end=174
  _globals['_PEERRECORD_ADDRESSINFO']._serialized_start=142
  _globals['_PEERRECORD_ADDRESSINFO']._serialized_end=174
# @@protoc_insertion_point(module_scope)



from nanolib.abc import IRawConnection
from nanolib.network.connection.raw_connection import RawConnection
from nanolib.security.pnet.psk_conn import PskConn


def new_protected_conn(conn: RawConnection | IRawConnection, psk: str) -> PskConn:
    if len(psk) != 64:
        raise ValueError("Expected 32-byte pre shared key (PSK)")

    return PskConn(conn, psk)



"""
QUIC Transport exceptions
"""

from typing import Any, Literal


class QUICError(Exception):
    """Base exception for all QUIC transport errors."""

    def __init__(self, message: str, error_code: int | None = None):
        super().__init__(message)
        self.error_code = error_code


# Transport-level exceptions


class QUICTransportError(QUICError):
    """Base exception for QUIC transport operations."""

    pass


class QUICDialError(QUICTransportError):
    """Error occurred during QUIC connection establishment."""

    pass


class QUICListenError(QUICTransportError):
    """Error occurred during QUIC listener operations."""

    pass


class QUICSecurityError(QUICTransportError):
    """Error related to QUIC security/TLS operations."""

    pass


# Connection-level exceptions


class QUICConnectionError(QUICError):
    """Base exception for QUIC connection operations."""

    pass


class QUICConnectionClosedError(QUICConnectionError):
    """QUIC connection has been closed."""

    pass


class QUICConnectionTimeoutError(QUICConnectionError):
    """QUIC connection operation timed out."""

    pass


class QUICHandshakeError(QUICConnectionError):
    """Error during QUIC handshake process."""

    pass


class QUICPeerVerificationError(QUICConnectionError):
    """Error verifying peer identity during handshake."""

    pass


# Stream-level exceptions


class QUICStreamError(QUICError):
    """Base exception for QUIC stream operations."""

    def __init__(
        self,
        message: str,
        stream_id: str | None = None,
        error_code: int | None = None,
    ):
        super().__init__(message, error_code)
        self.stream_id = stream_id


class QUICStreamClosedError(QUICStreamError):
    """Stream is closed and cannot be used for I/O operations."""

    pass


class QUICStreamResetError(QUICStreamError):
    """Stream was reset by local or remote peer."""

    def __init__(
        self,
        message: str,
        stream_id: str | None = None,
        error_code: int | None = None,
        reset_by_peer: bool = False,
    ):
        super().__init__(message, stream_id, error_code)
        self.reset_by_peer = reset_by_peer


class QUICStreamTimeoutError(QUICStreamError):
    """Stream operation timed out."""

    pass


class QUICStreamBackpressureError(QUICStreamError):
    """Stream write blocked due to flow control."""

    pass


class QUICStreamLimitError(QUICStreamError):
    """Stream limit reached (too many concurrent streams)."""

    pass


class QUICStreamStateError(QUICStreamError):
    """Invalid operation for current stream state."""

    def __init__(
        self,
        message: str,
        stream_id: str | None = None,
        current_state: str | None = None,
        attempted_operation: str | None = None,
    ):
        super().__init__(message, stream_id)
        self.current_state = current_state
        self.attempted_operation = attempted_operation


# Flow control exceptions


class QUICFlowControlError(QUICError):
    """Base exception for flow control related errors."""

    pass


class QUICFlowControlViolationError(QUICFlowControlError):
    """Flow control limits were violated."""

    pass


class QUICFlowControlDeadlockError(QUICFlowControlError):
    """Flow control deadlock detected."""

    pass


# Resource management exceptions


class QUICResourceError(QUICError):
    """Base exception for resource management errors."""

    pass


class QUICMemoryLimitError(QUICResourceError):
    """Memory limit exceeded."""

    pass


class QUICConnectionLimitError(QUICResourceError):
    """Connection limit exceeded."""

    pass


# Multiaddr and addressing exceptions


class QUICAddressError(QUICError):
    """Base exception for QUIC addressing errors."""

    pass


class QUICInvalidMultiaddrError(QUICAddressError):
    """Invalid multiaddr format for QUIC transport."""

    pass


class QUICAddressResolutionError(QUICAddressError):
    """Failed to resolve QUIC address."""

    pass


class QUICProtocolError(QUICError):
    """Base exception for QUIC protocol errors."""

    pass


class QUICVersionNegotiationError(QUICProtocolError):
    """QUIC version negotiation failed."""

    pass


class QUICUnsupportedVersionError(QUICProtocolError):
    """Unsupported QUIC version."""

    pass


# Configuration exceptions


class QUICConfigurationError(QUICError):
    """Base exception for QUIC configuration errors."""

    pass


class QUICInvalidConfigError(QUICConfigurationError):
    """Invalid QUIC configuration parameters."""

    pass


class QUICCertificateError(QUICConfigurationError):
    """Error with TLS certificate configuration."""

    pass


def map_quic_error_code(error_code: int) -> str:
    """
    Map QUIC error codes to human-readable descriptions.
    Based on RFC 9000 Transport Error Codes.
    """
    error_codes = {
        0x00: "NO_ERROR",
        0x01: "INTERNAL_ERROR",
        0x02: "CONNECTION_REFUSED",
        0x03: "FLOW_CONTROL_ERROR",
        0x04: "STREAM_LIMIT_ERROR",
        0x05: "STREAM_STATE_ERROR",
        0x06: "FINAL_SIZE_ERROR",
        0x07: "FRAME_ENCODING_ERROR",
        0x08: "TRANSPORT_PARAMETER_ERROR",
        0x09: "CONNECTION_ID_LIMIT_ERROR",
        0x0A: "PROTOCOL_VIOLATION",
        0x0B: "INVALID_TOKEN",
        0x0C: "APPLICATION_ERROR",
        0x0D: "CRYPTO_BUFFER_EXCEEDED",
        0x0E: "KEY_UPDATE_ERROR",
        0x0F: "AEAD_LIMIT_REACHED",
        0x10: "NO_VIABLE_PATH",
    }

    return error_codes.get(error_code, f"UNKNOWN_ERROR_{error_code:02X}")


def create_stream_error(
    error_type: str,
    message: str,
    stream_id: str | None = None,
    error_code: int | None = None,
) -> QUICStreamError:
    """
    Factory function to create appropriate stream error based on type.

    Args:
        error_type: Type of error ("closed", "reset", "timeout", "backpressure", etc.)
        message: Error message
        stream_id: Stream identifier
        error_code: QUIC error code

    Returns:
        Appropriate QUICStreamError subclass

    """
    error_type = error_type.lower()

    if error_type in ("closed", "close"):
        return QUICStreamClosedError(message, stream_id, error_code)
    elif error_type == "reset":
        return QUICStreamResetError(message, stream_id, error_code)
    elif error_type == "timeout":
        return QUICStreamTimeoutError(message, stream_id, error_code)
    elif error_type in ("backpressure", "flow_control"):
        return QUICStreamBackpressureError(message, stream_id, error_code)
    elif error_type in ("limit", "stream_limit"):
        return QUICStreamLimitError(message, stream_id, error_code)
    elif error_type == "state":
        return QUICStreamStateError(message, stream_id)
    else:
        return QUICStreamError(message, stream_id, error_code)


def create_connection_error(
    error_type: str, message: str, error_code: int | None = None
) -> QUICConnectionError:
    """
    Factory function to create appropriate connection error based on type.

    Args:
        error_type: Type of error ("closed", "timeout", "handshake", etc.)
        message: Error message
        error_code: QUIC error code

    Returns:
        Appropriate QUICConnectionError subclass

    """
    error_type = error_type.lower()

    if error_type in ("closed", "close"):
        return QUICConnectionClosedError(message, error_code)
    elif error_type == "timeout":
        return QUICConnectionTimeoutError(message, error_code)
    elif error_type == "handshake":
        return QUICHandshakeError(message, error_code)
    elif error_type in ("peer_verification", "verification"):
        return QUICPeerVerificationError(message, error_code)
    else:
        return QUICConnectionError(message, error_code)


class QUICErrorContext:
    """
    Context manager for handling QUIC errors with automatic error mapping.
    Useful for converting low-level aioquic errors to py-libp2p QUIC errors.
    """

    def __init__(self, operation: str, component: str = "quic") -> None:
        self.operation = operation
        self.component = component

    def __enter__(self) -> "QUICErrorContext":
        return self

    # TODO: Fix types for exc_type
    def __exit__(
        self,
        exc_type: type[BaseException] | None | None,
        exc_val: BaseException | None,
        exc_tb: Any,
    ) -> Literal[False]:
        if exc_type is None:
            return False

        if exc_val is None:
            return False

        # Map common aioquic exceptions to our exceptions
        if "ConnectionClosed" in str(exc_type):
            raise QUICConnectionClosedError(
                f"Connection closed during {self.operation}: {exc_val}"
            ) from exc_val
        elif "StreamReset" in str(exc_type):
            raise QUICStreamResetError(
                f"Stream reset during {self.operation}: {exc_val}"
            ) from exc_val
        elif "timeout" in str(exc_val).lower():
            if "stream" in self.component.lower():
                raise QUICStreamTimeoutError(
                    f"Timeout during {self.operation}: {exc_val}"
                ) from exc_val
            else:
                raise QUICConnectionTimeoutError(
                    f"Timeout during {self.operation}: {exc_val}"
                ) from exc_val
        elif "flow control" in str(exc_val).lower():
            raise QUICStreamBackpressureError(
                f"Flow control error during {self.operation}: {exc_val}"
            ) from exc_val

        # Let other exceptions propagate
        return False



from dataclasses import (
    dataclass,
)
import hmac

from Crypto.Cipher import (
    AES,
)
import Crypto.Util.Counter as Counter


class InvalidMACException(Exception):
    pass


@dataclass(frozen=True)
class EncryptionParameters:
    cipher_type: str
    hash_type: str
    iv: bytes
    mac_key: bytes
    cipher_key: bytes


class MacAndCipher:
    def __init__(self, parameters: EncryptionParameters) -> None:
        self.authenticator = hmac.new(
            parameters.mac_key, digestmod=parameters.hash_type
        )
        iv_bit_size = 8 * len(parameters.iv)
        cipher = AES.new(
            parameters.cipher_key,
            AES.MODE_CTR,
            counter=Counter.new(
                iv_bit_size,
                initial_value=int.from_bytes(parameters.iv, byteorder="big"),
            ),
        )
        self.cipher = cipher

    def encrypt(self, data: bytes) -> bytes:
        return self.cipher.encrypt(data)

    def authenticate(self, data: bytes) -> bytes:
        authenticator = self.authenticator.copy()
        authenticator.update(data)
        return authenticator.digest()

    def decrypt_if_valid(self, data_with_tag: bytes) -> bytes:
        tag_position = len(data_with_tag) - self.authenticator.digest_size
        data = data_with_tag[:tag_position]
        tag = data_with_tag[tag_position:]

        authenticator = self.authenticator.copy()
        authenticator.update(data)
        expected_tag = authenticator.digest()

        if not hmac.compare_digest(tag, expected_tag):
            raise InvalidMACException(expected_tag, tag)

        return self.cipher.decrypt(data)


def initialize_pair(
    cipher_type: str, hash_type: str, secret: bytes
) -> tuple[EncryptionParameters, EncryptionParameters]:
    """
    Return a pair of ``Keys`` for use in securing a communications channel
    with authenticated encryption derived from the ``secret`` and using the
    requested ``cipher_type`` and ``hash_type``.
    """
    if cipher_type != "AES-128":
        raise NotImplementedError()
    if hash_type != "SHA256":
        raise NotImplementedError()

    iv_size = 16
    cipher_key_size = 16
    hmac_key_size = 20
    seed = b"key expansion"

    params_size = iv_size + cipher_key_size + hmac_key_size
    result = bytearray(2 * params_size)

    authenticator = hmac.new(secret, digestmod=hash_type)
    authenticator.update(seed)
    tag = authenticator.digest()

    i = 0
    len_result = 2 * params_size
    while i < len_result:
        authenticator = hmac.new(secret, digestmod=hash_type)

        authenticator.update(tag)
        authenticator.update(seed)

        another_tag = authenticator.digest()

        remaining_bytes = len(another_tag)

        if i + remaining_bytes > len_result:
            remaining_bytes = len_result - i

        result[i : i + remaining_bytes] = another_tag[0:remaining_bytes]

        i += remaining_bytes

        authenticator = hmac.new(secret, digestmod=hash_type)
        authenticator.update(tag)
        tag = authenticator.digest()

    first_half = result[:params_size]
    second_half = result[params_size:]

    return (
        EncryptionParameters(
            cipher_type,
            hash_type,
            bytes(first_half[0:iv_size]),
            bytes(first_half[iv_size + cipher_key_size :]),
            bytes(first_half[iv_size : iv_size + cipher_key_size]),
        ),
        EncryptionParameters(
            cipher_type,
            hash_type,
            bytes(second_half[0:iv_size]),
            bytes(second_half[iv_size + cipher_key_size :]),
            bytes(second_half[iv_size : iv_size + cipher_key_size]),
        ),
    )



"""
Prometheus metrics exporter for libp2p resource manager.
"""

from __future__ import annotations

import threading

try:
    from prometheus_client import (  # type: ignore[import]
        CONTENT_TYPE_LATEST,
        CollectorRegistry,
        Gauge,
        Histogram,
        generate_latest,
        start_http_server,
    )

    PROMETHEUS_AVAILABLE = True
except ImportError:
    PROMETHEUS_AVAILABLE = False

from .metrics import Metrics


class PrometheusExporter:
    """
    Exports libp2p resource manager metrics in Prometheus format.

    This exporter provides metrics compatible with the go-libp2p resource manager
    format, allowing the use of existing Grafana dashboards and monitoring tools.
    """

    def __init__(
        self,
        port: int = 8000,
        registry: CollectorRegistry | None = None,
        enable_server: bool = True,
    ) -> None:
        """
        Initialize the Prometheus exporter.

        Args:
            port: HTTP port for metrics endpoint
            registry: Custom Prometheus registry (uses default if None)
            enable_server: Whether to start the HTTP server automatically

        Raises:
            ImportError: If prometheus_client is not available

        """
        if not PROMETHEUS_AVAILABLE:
            raise ImportError(
                "prometheus_client is required for Prometheus metrics export. "
                "Install with: pip install prometheus-client"
            )

        self.port = port
        self.registry = registry or CollectorRegistry()
        self.enable_server = enable_server
        self._server_thread: threading.Thread | None = None
        self._running = False
        self._lock = threading.RLock()

        # Peer tracking for histogram metrics
        self._peer_connections: dict[str, dict[str, int]] = {}
        self._peer_streams: dict[str, dict[str, int]] = {}
        self._peer_memory: dict[str, int] = {}
        self._conn_memory: dict[str, int] = {}

        # Declare metric attributes
        self.connections: Gauge
        self.streams: Gauge
        self.memory: Gauge
        self.fds: Gauge
        self.blocked_resources: Gauge
        self.peer_connections: Histogram
        self.previous_peer_connections: Histogram
        self.peer_streams: Histogram
        self.previous_peer_streams: Histogram
        self.peer_memory: Histogram
        self.previous_peer_memory: Histogram
        self.conn_memory: Histogram
        self.previous_conn_memory: Histogram

        # Initialize Prometheus metrics (compatible with go-libp2p)
        self._init_metrics()

        if enable_server:
            self.start_server()

    def _init_metrics(self) -> None:
        """Initialize Prometheus metrics compatible with go-libp2p format."""
        # Connection metrics - matches go-libp2p format exactly
        self.connections = Gauge(
            "libp2p_rcmgr_connections",
            "Number of connections",
            ["dir", "scope"],
            registry=self.registry,
        )

        # Stream metrics - matches go-libp2p format
        self.streams = Gauge(
            "libp2p_rcmgr_streams",
            "Number of streams",
            ["dir", "scope", "protocol"],
            registry=self.registry,
        )

        # Memory metrics - matches go-libp2p format
        self.memory = Gauge(
            "libp2p_rcmgr_memory",
            "Memory usage as reported to the Resource Manager",
            ["scope", "protocol"],
            registry=self.registry,
        )

        # File descriptor metrics
        self.fds = Gauge(
            "libp2p_rcmgr_fds",
            "File descriptors reserved as reported to the Resource Manager",
            ["scope"],
            registry=self.registry,
        )

        # Blocked resources counter - matches go-libp2p format
        self.blocked_resources = Gauge(
            "libp2p_rcmgr_blocked_resources",
            "Number of blocked resources",
            ["dir", "scope", "resource"],
            registry=self.registry,
        )

        # Peer connection distribution - go-libp2p compatible buckets
        connection_buckets = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 16, 32, 64, 128, 256]

        self.peer_connections = Histogram(
            "libp2p_rcmgr_peer_connections",
            "Number of connections this peer has",
            ["dir"],
            buckets=connection_buckets,
            registry=self.registry,
        )

        # Previous peer connections for histogram diff calculation
        self.previous_peer_connections = Histogram(
            "libp2p_rcmgr_previous_peer_connections",
            "Number of connections this peer previously had",
            ["dir"],
            buckets=connection_buckets,
            registry=self.registry,
        )

        # Peer stream distribution
        self.peer_streams = Histogram(
            "libp2p_rcmgr_peer_streams",
            "Number of streams this peer has",
            ["dir"],
            buckets=connection_buckets,  # Same buckets as connections
            registry=self.registry,
        )

        # Previous peer streams for histogram diff calculation
        self.previous_peer_streams = Histogram(
            "libp2p_rcmgr_previous_peer_streams",
            "Number of streams this peer previously had",
            ["dir"],
            buckets=connection_buckets,
            registry=self.registry,
        )

        # Memory distribution buckets (bytes: 1KB to 4GB)
        mem_buckets = [
            1024,  # 1KB
            4096,  # 4KB
            32768,  # 32KB
            1048576,  # 1MB
            33554432,  # 32MB
            268435456,  # 256MB
            536870912,  # 512MB
            1073741824,  # 1GB
            2147483648,  # 2GB
            4294967296,  # 4GB
        ]

        self.peer_memory = Histogram(
            "libp2p_rcmgr_peer_memory",
            "How many peers have reserved this bucket of memory",
            buckets=mem_buckets,
            registry=self.registry,
        )

        self.previous_peer_memory = Histogram(
            "libp2p_rcmgr_previous_peer_memory",
            "How many peers have previously reserved this bucket of memory",
            buckets=mem_buckets,
            registry=self.registry,
        )

        self.conn_memory = Histogram(
            "libp2p_rcmgr_conn_memory",
            "How many connections have reserved this bucket of memory",
            buckets=mem_buckets,
            registry=self.registry,
        )

        self.previous_conn_memory = Histogram(
            "libp2p_rcmgr_previous_conn_memory",
            "How many connections have previously reserved this bucket of memory",
            buckets=mem_buckets,
            registry=self.registry,
        )

    def update_from_metrics(self, metrics: Metrics) -> None:
        """
        Update Prometheus metrics from libp2p Metrics instance.

        Args:
            metrics: The metrics instance to export data from

        """
        with self._lock:
            summary = metrics.get_summary()

            # Update connection metrics
            self.connections.labels(dir="inbound", scope="system").set(
                summary["connections"]["inbound"]
            )
            self.connections.labels(dir="outbound", scope="system").set(
                summary["connections"]["outbound"]
            )

            # Update stream metrics (system scope)
            self.streams.labels(dir="inbound", scope="system", protocol="").set(
                summary["streams"]["inbound"]
            )
            self.streams.labels(dir="outbound", scope="system", protocol="").set(
                summary["streams"]["outbound"]
            )

            # Update memory metrics
            self.memory.labels(scope="system", protocol="").set(
                summary["memory"]["current"]
            )

            # Update file descriptors (estimate based on connections)
            # In a real implementation, you'd track FDs separately
            total_connections = summary["connections"]["total"]
            self.fds.labels(scope="system").set(total_connections)

            # Update blocked resources (convert counters to gauges for compatibility)
            self.blocked_resources.labels(
                dir="", scope="system", resource="connection"
            ).set(summary["blocks"]["connections"])
            self.blocked_resources.labels(
                dir="", scope="system", resource="stream"
            ).set(summary["blocks"]["streams"])
            self.blocked_resources.labels(
                dir="", scope="system", resource="memory"
            ).set(summary["blocks"]["memory"])

    def record_blocked_resource(
        self,
        direction: str,
        scope: str,
        resource: str,
        count: int = 1,
    ) -> None:
        """
        Record a blocked resource event.

        Args:
            direction: Direction of the blocked resource (inbound/outbound)
            scope: Resource scope (system, transient, peer, etc.)
            resource: Type of resource (connection, stream, memory, fd)
            count: Number of resources blocked

        """
        with self._lock:
            current = self.blocked_resources.labels(
                dir=direction, scope=scope, resource=resource
            )._value._value
            self.blocked_resources.labels(
                dir=direction, scope=scope, resource=resource
            ).set(current + count)

    def record_peer_connections(
        self,
        peer_id: str,
        direction: str,
        old_count: int,
        new_count: int,
    ) -> None:
        """
        Record peer connection count for histogram (go-libp2p compatible).

        Args:
            peer_id: Peer identifier
            direction: Connection direction (inbound/outbound)
            old_count: Previous connection count
            new_count: Current connection count

        """
        with self._lock:
            if old_count != new_count:
                if old_count > 0:
                    self.previous_peer_connections.labels(dir=direction).observe(
                        old_count
                    )
                if new_count > 0:
                    self.peer_connections.labels(dir=direction).observe(new_count)

            # Update tracking
            if peer_id not in self._peer_connections:
                self._peer_connections[peer_id] = {"inbound": 0, "outbound": 0}
            self._peer_connections[peer_id][direction] = new_count

    def record_peer_streams(
        self,
        peer_id: str,
        direction: str,
        old_count: int,
        new_count: int,
    ) -> None:
        """
        Record peer stream count for histogram (go-libp2p compatible).

        Args:
            peer_id: Peer identifier
            direction: Stream direction (inbound/outbound)
            old_count: Previous stream count
            new_count: Current stream count

        """
        with self._lock:
            if old_count != new_count:
                if old_count > 0:
                    self.previous_peer_streams.labels(dir=direction).observe(old_count)
                if new_count > 0:
                    self.peer_streams.labels(dir=direction).observe(new_count)

            # Update tracking
            if peer_id not in self._peer_streams:
                self._peer_streams[peer_id] = {"inbound": 0, "outbound": 0}
            self._peer_streams[peer_id][direction] = new_count

    def record_peer_memory(self, peer_id: str, old_bytes: int, new_bytes: int) -> None:
        """
        Record peer memory usage for histogram.

        Args:
            peer_id: Peer identifier
            old_bytes: Previous memory usage in bytes
            new_bytes: Current memory usage in bytes

        """
        with self._lock:
            if old_bytes != new_bytes:
                if old_bytes > 0:
                    self.previous_peer_memory.observe(old_bytes)
                if new_bytes > 0:
                    self.peer_memory.observe(new_bytes)

            self._peer_memory[peer_id] = new_bytes

    def record_conn_memory(self, conn_id: str, old_bytes: int, new_bytes: int) -> None:
        """
        Record connection memory usage for histogram.

        Args:
            conn_id: Connection identifier
            old_bytes: Previous memory usage in bytes
            new_bytes: Current memory usage in bytes

        """
        with self._lock:
            if old_bytes != new_bytes:
                if old_bytes > 0:
                    self.previous_conn_memory.observe(old_bytes)
                if new_bytes > 0:
                    self.conn_memory.observe(new_bytes)

            self._conn_memory[conn_id] = new_bytes

    def start_server(self) -> None:
        """Start the Prometheus metrics HTTP server."""
        if self._running or not self.enable_server:
            return

        self._running = True
        self._server_thread = threading.Thread(
            target=self._run_server, daemon=True, name="PrometheusMetricsServer"
        )
        self._server_thread.start()

    def _run_server(self) -> None:
        """Run the Prometheus HTTP server."""
        try:
            start_http_server(self.port, registry=self.registry)
        except Exception as e:
            print(f"Failed to start Prometheus metrics server on port {self.port}: {e}")
            self._running = False

    def stop_server(self) -> None:
        """Stop the Prometheus metrics server."""
        self._running = False

    def get_metrics_text(self) -> str:
        """
        Get metrics in Prometheus text format.

        Returns:
            Metrics in Prometheus exposition format

        """
        return generate_latest(self.registry).decode("utf-8")

    def get_metrics_content_type(self) -> str:
        """
        Get the content type for Prometheus metrics.

        Returns:
            Content type string

        """
        return CONTENT_TYPE_LATEST

    def reset(self) -> None:
        """Reset all tracking data."""
        with self._lock:
            self._peer_connections.clear()
            self._peer_streams.clear()
            self._peer_memory.clear()
            self._conn_memory.clear()


def create_prometheus_exporter(
    port: int = 8000,
    enable_server: bool = True,
) -> PrometheusExporter | None:
    """
    Create a PrometheusExporter instance if prometheus_client is available.

    Args:
        port: HTTP port for metrics endpoint
        enable_server: Whether to start the HTTP server automatically

    Returns:
        PrometheusExporter instance or None if prometheus_client not available

    """
    if not PROMETHEUS_AVAILABLE:
        print(
            "Warning: prometheus_client not available. "
            "Install with: pip install prometheus-client"
        )
        return None

    try:
        return PrometheusExporter(port=port, enable_server=enable_server)
    except Exception as e:
        print(f"Failed to create Prometheus exporter: {e}")
        return None
